# https://github.com/AndrewCarterUK/pascal-voc-writer
import os
import shutil
import traceback
import tempfile
import logging
from PIL import Image
import json
from jinja2 import Environment, PackageLoader
from concurrent.futures import ThreadPoolExecutor
import dtlpy as dl

logger = logging.getLogger(name='dtlpy')


class BaseConverterFromPlatform:
    def __init__(self, project_name, dataset_name, output_directory, remote_path):
        self.name = ''
        self.project_name = project_name
        self.dataset_name = dataset_name
        self.remote_path = remote_path
        self.output_directory = output_directory
        self.params = None

        # for threading outputs
        self.errors = None
        self.outputs = None
        self.results = None

    def convert_single_file(self, output_directory, item, annotations, params):
        # for user to create
        pass

    def threading_wrapper(self, func, i_item,
                          # inputs for user
                          output_directory, item, annotations, params=None):
        try:
            self.outputs[i_item] = func(output_directory=output_directory,
                                        item=item,
                                        annotations=annotations,
                                        params=params)
            self.results[i_item] = True
        except Exception as err:
            logging.exception('Error running thread conversion')
            self.errors[i_item] = traceback.format_exc()
            self.results[i_item] = False

    def run(self):
        # create temp path to save dataloop annotations
        local_annotations_path = os.path.join(tempfile.gettempdir(),
                                              'dataloop_annotations_{}'.format(hash(os.times())))
        if os.path.isdir(local_annotations_path):
            raise IsADirectoryError('path already exists')

        try:
            # download annotations zip to local directory
            project = dl.projects.get(project_name=self.project_name)
            dataset = project.datasets.get(dataset_name=self.dataset_name)

            dataset.items.download_annotations(dataset_name=self.dataset_name,
                                               local_path=os.path.join(local_annotations_path, '*'))

            # get labels to ids dictionary
            if 'labels_dict' not in self.params:
                self.params['labels_dict'] = {label: i_label for i_label, label in
                                              enumerate(list(dataset.labels.keys()))}

            output_annotations_path = os.path.join(self.output_directory, 'annotations')
            # create output directories
            if not os.path.isdir(self.output_directory):
                os.makedirs(self.output_directory)
            if not os.path.isdir(output_annotations_path):
                os.makedirs(output_annotations_path)

            # save labels
            with open(os.path.join(self.output_directory, 'labels.txt'), 'w') as f:
                f.write('\n'.join(['%s:%s' % (val, key) for key, val in self.params['labels_dict'].items()]))

            # get all items (for width and height)
            filters = dl.Filters()
            filters.add(field='filename', values=self.remote_path)
            filters.add(field='type', values='file')
            pages = dataset.items.list(filters=filters)

            # init workers and results lists
            pool = ThreadPoolExecutor(max_workers=32)
            i_item = -1
            num_items = pages.items_count
            self.outputs = [None for _ in range(num_items)]
            self.results = [None for _ in range(num_items)]
            self.errors = [None for _ in range(num_items)]

            for page in pages:
                for item in page:
                    i_item += 1
                    # create input annotations json
                    in_filepath = os.path.join(local_annotations_path, item.filename[1:])
                    name, ext = os.path.splitext(in_filepath)
                    in_filepath = name + '.json'

                    # check if annotations file exists
                    if not os.path.isfile(in_filepath):
                        self.results[i_item] = False
                        self.errors[i_item] = 'file not found: %s' % in_filepath
                        continue

                    with open(in_filepath, 'r', encoding="utf8") as f:
                        data = json.load(f)

                    pool.submit(self.threading_wrapper, **{'func': self.convert_single_file,
                                                           'i_item': i_item,
                                                           # input for "func"
                                                           'output_directory': output_annotations_path,
                                                           'item': item,
                                                           'annotations': data['annotations'],
                                                           'params': self.params})
            print('Done')
            pool.shutdown()
            dummy = [logger.error(self.errors[i_job]) for i_job, suc in enumerate(self.results) if suc is False]
            return self.outputs
        except:
            raise
        finally:
            # cleanup
            if os.path.isdir(local_annotations_path):
                shutil.rmtree(local_annotations_path)


class DtlpyToVoc(BaseConverterFromPlatform):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # specific for voc
        labels = dict()
        # annotations template
        environment = Environment(loader=PackageLoader('dtlpy', 'assets'),
                                  keep_trailing_newline=True)
        annotation_template = environment.get_template('voc_annotation_template.xml')
        self.params = {'labels': labels,
                       'annotation_template': annotation_template}

    @staticmethod
    def new_annotation(path, width, height, depth=3, database='Unknown', segmented=0):
        abspath = os.path.abspath(path)
        annotation = {
            'path': abspath,
            'filename': os.path.basename(abspath),
            'folder': os.path.basename(os.path.dirname(abspath)),
            'width': width,
            'height': height,
            'depth': depth,
            'database': database,
            'segmented': segmented,
            'objects': list()
        }
        return annotation

    def convert_single_file(self, item, annotations, output_directory, params):
        # output filepath for xml
        out_filepath = os.path.join(output_directory, item.filename[1:])
        # remove ext from output filepath
        out_filepath, ext = os.path.splitext(out_filepath)
        # add xml extension
        out_filepath += '.xml'
        if not os.path.isdir(os.path.dirname(out_filepath)):
            os.makedirs(os.path.dirname(out_filepath), exist_ok=True)

        width = item.width
        height = item.height
        depth = item.metadata['system'].get('channels', 3)
        output_annotation = {
            'path': item.filename,
            'filename': os.path.basename(item.filename),
            'folder': os.path.basename(os.path.dirname(item.filename)),
            'width': width,
            'height': height,
            'depth': depth,
            'database': 'Unknown',
            'segmented': 0,
            'objects': list()
        }

        for annotation in annotations:
            if not annotation:
                continue
            if annotation['type'] != 'box':
                continue
            label = annotation['label']
            coordinates = annotation['coordinates']

            attributes = list()
            # get attributes if exists
            if 'attributes' in annotation:
                attributes = annotation['attributes']

            try:
                left = int(coordinates[0]['x'])
                top = int(coordinates[0]['y'])
                right = int(coordinates[1]['x'])
                bottom = int(coordinates[1]['y'])
            except Exception as err:
                print('coordinates', coordinates)
                continue

            output_annotation['objects'].append({'name': label,
                                                 'xmin': left,
                                                 'ymin': top,
                                                 'xmax': right,
                                                 'ymax': bottom,
                                                 'attributes': attributes,
                                                 })
        with open(out_filepath, 'w') as file:
            content = params['annotation_template'].render(**output_annotation)
            file.write(content)


class DtlpyToYolo(BaseConverterFromPlatform):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.params = {'labels': dict()}

    @staticmethod
    def convert_bb(size, box):
        dw = 1. / size[0]
        dh = 1. / size[1]
        x = (box[0] + box[1]) / 2.0
        y = (box[2] + box[3]) / 2.0
        w = box[1] - box[0]
        h = box[3] - box[2]
        x = x * dw
        w = w * dw
        y = y * dh
        h = h * dh
        return x, y, w, h

    @staticmethod
    def convert_single_file(item, annotations, output_directory, params):
        # output filepath for xml
        out_filepath = os.path.join(output_directory, item.filename[1:])
        # remove ext from filepath
        out_filepath, ext = os.path.splitext(out_filepath)
        # add txt extension
        out_filepath += '.txt'

        if not os.path.isdir(os.path.dirname(out_filepath)):
            os.makedirs(os.path.dirname(out_filepath), exist_ok=True)

        width = item.width
        height = item.height
        yolo_annotations = list()
        for annotation in annotations:
            if annotation['type'] != 'box':
                continue
            label = annotation['label']
            coordinates = annotation['coordinates']
            try:
                left = int(coordinates[0]['x'])
                top = int(coordinates[0]['y'])
                right = int(coordinates[1]['x'])
                bottom = int(coordinates[1]['y'])
            except Exception as err:
                print('coords', coordinates)
                continue
            yolo_bb = DtlpyToYolo.convert_bb((width, height), (left, right, top, bottom))
            yolo_annotations.append(
                '%d %f %f %f %f' % (params['labels_dict'][label], yolo_bb[0], yolo_bb[1], yolo_bb[2], yolo_bb[3]))

        with open(out_filepath, 'w') as f:
            f.write('\n'.join(yolo_annotations))



import dtlpy as dl
import logging

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """

    def run(self, annotation, progress=None):
        """
        In this example we update item's metadata if annotation id of specific type

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        assert isinstance(annotation, dl.Annotation)
        if annotation.label == 'box' and annotation.label == 'dog':
            logger.info('Dog was detected in item')
            annotation.attributes.append('Dog detection')
            annotation.item.metadata['dogInPicture'] = True
            annotation.update()
            annotation.item.update()
        logger.info('Function finished successfully!')


if __name__ == "__main__":
    """
    Run this main to locally debug your package
    """
    dl.packages.test_local_package()



import os
import json
import numpy as np
import dtlpy as dl
import tempfile


class Report:
    def __init__(self, nrows: int = 1, ncols: int = 1):
        """
        Create a report layout with defined size
        :param int nrows: number of rows in the report json
        :param int ncols: number of columns in the report json
        """
        self.figs = np.empty(shape=(nrows, ncols), dtype=object)
        self._ncols = ncols
        self._nrows = nrows
        self._irow = 0
        self._icol = 0

    def add(self, fig, irow: int, icol: int):
        """
        Add fig to to an index location in the layout

        :param fig: input figure.
        :param int irow: row location of the added fig. should be within the layout boundaries defined in the Report init
        :param int icol: column location of the added fig. should be within the layout boundaries defined in the Report init
        :return:
        """
        if 0 > irow or irow >= self._nrows:
            raise ValueError(
                'Trying to set irow with value outside layout definition. layout rows: {}'.format(self._nrows))
        if 0 > icol or icol >= self._ncols:
            raise ValueError(
                'Trying to set icol with value outside layout definition. layout cols: {}'.format(self._ncols))
        self.figs[irow, icol] = fig.to_dict()

    def prepare(self):
        """
        Covert the layout into the json item for uploading

        :return:
        """
        return {
            "shebang": "dataloop",
            "metadata": {"dltype": "report"},
            "layout": {
                "rows": [{"cols": [fig for fig in row if fig is not None]} for row in self.figs]
                # "rows": [{"cols": [report_json]}]
            }
        }

    def upload(self, dataset: dl.Dataset, remote_name: str, remote_path: str):
        """
        Upload the report item to the platform

        :param dataset: dl.Dataset to upload the item to
        :param remote_name: name of the report. should be with .json extension
        :param remote_path: remote directory to upload the item to
        :return:
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            filepath = os.path.join(tmpdir, remote_name)
            print(filepath)
            with open(filepath, 'w') as f:
                json.dump(self.prepare(), f)
            item = dataset.items.upload(local_path=filepath,
                                        remote_path=remote_path,
                                        remote_name=remote_name,
                                        overwrite=True)
        return item



import numpy as np

from . import BaseAnnotationDefinition


class Point(BaseAnnotationDefinition):
    """
    Point annotation object
    """
    type = "point"

    def __init__(self, x, y, label, attributes=None, description=None):
        super().__init__(description=description, attributes=attributes)
        self.y = y
        self.x = x
        self.label = label

    @property
    def geo(self):
        return [self.x, self.y]

    @property
    def left(self):
        return self.x

    @property
    def top(self):
        return self.y

    @property
    def right(self):
        return self.y

    @property
    def bottom(self):
        return self.x

    @property
    def z(self):
        return 0

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        # point cant have thickness 1
        if thickness is None or thickness == -1:
            thickness = 5

        # create image to draw on
        if alpha != 1:
            overlay = image.copy()
        else:
            overlay = image

        # draw annotation
        overlay = cv2.circle(
            img=overlay,
            center=(int(np.round(self.x)), int(np.round(self.y))),
            radius=thickness,
            color=color,
            thickness=thickness,
            lineType=cv2.LINE_AA,
        )

        if not isinstance(color, int) and len(color) == 4 and color[3] != 255:
            # add with opacity
            image = cv2.addWeighted(src1=overlay,
                                    alpha=alpha,
                                    src2=image,
                                    beta=1 - alpha,
                                    gamma=0)
        else:
            image = overlay

        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    def to_coordinates(self, color):
        return {"x": float(self.x), "y": float(self.y), "z": float(self.z)}

    @staticmethod
    def from_coordinates(coordinates):
        return coordinates["x"], coordinates["y"]

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            x, y = cls.from_coordinates(_json["coordinates"])
        elif "data" in _json:
            x, y = cls.from_coordinates(_json["data"])
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))

        return cls(
            x=x,
            y=y,
            label=_json["label"],
            attributes=_json.get("attributes", None),
        )



from . import add_labels
from . import add_metadata_to_item
from . import annotate_items_using_model
from . import annotate_video_using_model_and_tracker
from . import annotations_convert_to_voc
from . import annotations_convert_to_yolo
from . import convert_annotation_types
from . import copy_annotations
from . import copy_folder
from . import create_annotations
from . import create_video_annotations
from . import move_item
from . import play_video_annotation
from . import show_item_and_mask
from . import upload_batch_of_items
from . import upload_items_and_custom_format_annotations
from . import upload_segmentation_annotations_from_mask_image
from . import upload_yolo_format_annotations
from . import upload_items_with_modalities



import attr
import logging
import random

from .. import PlatformException

logger = logging.getLogger(name='dtlpy')


@attr.s
class Label:
    tag = attr.ib()
    display_data = attr.ib()
    color = attr.ib(default=None)
    display_label = attr.ib(default=None)
    attributes = attr.ib()
    children = attr.ib()

    @attributes.default
    def set_attributes(self):
        attributes = list()
        return attributes

    @children.default
    def set_children(self):
        children = list()
        return children


    @display_data.default
    def set_display_data(self):
        display_data = dict()
        return display_data

    @classmethod
    def from_root(cls, root):
        """
        Build a Label entity object from a json

        :param dict root: _json representation of a label as it is in host
        :return: Label object
        """
        children = list()
        if 'children' in root and root['children'] is not None:
            children = [Label.from_root(child) for child in root['children']]

        root = root.get("value", root)
        if "tag" in root:
            label_name = root["tag"]
        elif "label_name" in root:
            label_name = root["label_name"]
        else:
            raise PlatformException("400", "Invalid input - each label must have a tag")

        display_label = root.get("displayLabel", None)
        if display_label is None:
            display_label = root.get("display_label", None)

        display_data = root.get("displayData", dict())
        return cls(
            tag=label_name,
            display_data=display_data,
            color=root.get("color", None),
            display_label=display_label,
            attributes=root.get("attributes", None),
            children=children
        )

    def to_root(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        value = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Label).children,
                                                              attr.fields(Label).color,
                                                              attr.fields(Label).display_label,
                                                              attr.fields(Label).display_data))
        value['displayLabel'] = self.display_label
        value['displayData'] = self.display_data
        if self.color:
            value['color'] = self.hex
        children = [child.to_root() for child in self.children]
        _json = {
            'value': value,
            'children': children
        }
        return _json

    @property
    def rgb(self):
        """
        Return label's color in RBG format

        :return: label's color in RBG format
        """
        if self.color is None:
            color = None
        elif isinstance(self.color, str) and self.color.startswith('rgb'):
            color = tuple(eval(self.color.lstrip('rgb')))
        elif isinstance(self.color, str) and self.color.startswith('#'):
            color = tuple(int(self.color.lstrip('#')[i:i + 2], 16) for i in (0, 2, 4))
        elif isinstance(self.color, tuple) or isinstance(self.color, list):
            color = self.color
        else:
            logger.warning('Unknown color scheme: {}'.format(self.color))
            color = (255, 0, 0)
        return color

    @property
    def hex(self):
        """
        Return label's color in HEX format

        :return: label's color in HEX format
        """
        if isinstance(self.color, tuple) or isinstance(self.color, list):
            return '#%02x%02x%02x' % self.color
        elif self.color.startswith('rgb'):
            rgb = tuple(eval(self.color.lstrip('rgb')))
            return '#%02x%02x%02x' % rgb
        elif self.color.startswith('#'):
            return self.color



import importlib.util
import inspect
import logging
import typing

from .. import entities

logger = logging.getLogger(name='dtlpy')


class PackageModule(entities.DlEntity):
    """
    PackageModule object
    """
    # platform
    name: str = entities.DlProperty(location=['name'], _type=str)
    init_inputs: typing.List['entities.FunctionIO'] = entities.DlProperty(location=['initInputs'],
                                                                          _type=typing.Union[list, None],
                                                                          _kls='FunctionIO')

    entry_point = entities.DlProperty(location=['entryPoint'],
                                      _type=str,
                                      default=entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT)
    class_name = entities.DlProperty(location=['className'],
                                     _type=dict,
                                     default=entities.package_defaults.DEFAULT_PACKAGE_CLASS_NAME)
    functions: typing.List['entities.PackageFunction'] = entities.DlProperty(location=['functions'],
                                                                             _type=list,
                                                                             default=list(),
                                                                             _kls='PackageFunction')
    compute_config: str = entities.DlProperty(location=['computeConfig'], _type=str, default=None)

    def __repr__(self):
        # TODO need to move to DlEntity
        return f"PackageModule(name={self.name}, entry_point={self.entry_point}, class_name={self.class_name})"

    @functions.validator
    def validate_functions(self, value: list):
        if not isinstance(value, list):
            raise Exception('Module functions must be a list.')
        if not self.unique_functions(value):
            raise Exception('Cannot have 2 functions by the same name in one module.')

    @staticmethod
    def unique_functions(functions: list):
        return len(functions) == len(set([function.name for function in functions]))

    @name.default
    def set_name(self):
        logger.warning('No module name was given. Using default name: {}'.format(
            entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME))
        return entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME

    @init_inputs.default
    def set_init_inputs(self):
        return list()

    @classmethod
    def from_json(cls, _json):
        inst = cls(_dict=_json)
        return inst

    @classmethod
    def from_entry_point(cls, entry_point):
        """
        Create a dl.PackageModule entity using decorator on the service class.

        :param entry_point: path to the python file with the runner class (relative to the package path)
        :return:
        """
        file_spec = importlib.util.spec_from_file_location(entry_point, entry_point)
        file_module = importlib.util.module_from_spec(file_spec)
        file_spec.loader.exec_module(file_module)
        module = None
        for cls_name, cls_inst in inspect.getmembers(file_module, predicate=inspect.isclass):
            spec = getattr(cls_inst, '__dtlpy__', None)
            if spec is not None:
                functions = spec['functions']
                available_methods = [name for name in ['train', 'predict']
                                     if 'BaseModelAdapter' not in getattr(cls_inst, name).__qualname__]
                if "train" not in available_methods:
                    # remove train_model from functions list if train is not available
                    functions[:] = [d for d in functions if d.get('name') != "train_model"]
                if "predict" not in available_methods:
                    # remove predict_items from functions list if predict is not available
                    functions[:] = [d for d in functions if d.get('name') != "predict_items"]
                if "extract_features" not in available_methods:
                    # remove extract_item_features from functions list if extract_features is not available
                    functions[:] = [d for d in functions if d.get('name') != "extract_item_features"]
                spec['entryPoint'] = entry_point
                spec['functions'] = functions
                module = cls.from_json(spec)
                break
        if module is None:
            raise ValueError('Failed to find a decorated Runner class in file: {}'.format(entry_point))
        return module

    def add_function(self, function):
        """
        :param function:
        """
        if not isinstance(self.functions, list):
            self.functions = [self.functions]
        if isinstance(function, entities.PackageFunction):
            self.functions.append(function)
        elif isinstance(function, dict):
            self.functions.append(entities.PackageFunction.from_json(function))
        else:
            raise ValueError('Unknown function type: {}. Expecting dl.PackageFunction or dict')

    def to_json(self):
        _json = self._dict.copy()
        return _json



"""

"""


def main():
    from PIL import Image
    import numpy as np
    import dtlpy as dl

    # Get project and dataset
    project = dl.projects.get(project_name='Presidents')
    dataset = project.datasets.get(dataset_name='William Henry Harrison')

    # image filepath
    image_filepath = '/home/images/with_family.png'
    # annotations filepath - RGB with color for each label
    annotations_filepath = '/home/masks/with_family.png'

    # upload item to root directory
    item = dataset.items.upload(local_path=image_filepath,
                                remote_path='/')

    # read mask from file
    mask = np.array(Image.open(annotations_filepath))

    # get unique color (labels)
    unique_colors = np.unique(mask.reshape(-1, mask.shape[2]), axis=0)

    # init dataloop annotations builder
    builder = item.annotations.builder()
    # for each label - create a dataloop mask annotation
    for i, color in enumerate(unique_colors):
        print(color)
        if i == 0:
            # ignore background
            continue
        # get mask of same color
        class_mask = np.all(color == mask, axis=2)
        # add annotation to builder
        builder.add(annotation_definition=dl.Segmentation(geo=class_mask,
                                                          label=str(i)))
    # upload all annotations
    item.annotations.upload(builder)



import redis
import json
import datetime
from .base_cache import BaseCache


class RedisCache(BaseCache):
    def __init__(self, options=None, ttl=1000):
        if options is None:
            options = {}
        self.cache = redis.Redis(
            host=options.get('host', '127.0.0.1'),
            port=options.get('port', 6379),
            socket_keepalive=options.get('keepalive', True)
        )
        self.ttl = ttl

    def set(self, key, value):
        """
        set or add a key and value to the cache
        :param key: str or int type of key
        :param value: pickled value
        :return:
        """
        if not isinstance(key, str):
            raise ValueError("key must be string")
        self.cache.setex(name=key, value=value, time=self.ttl)

    def ping(self):
        """
        Cache ping check if connection is working
        """
        self.cache.ping()

    def list(self, pattern):
        """
        get the value of the key from the cache
        :param pattern: str or int type of key
        :return: the value of the key
        """
        if '\\' in pattern:
            pattern = pattern.replace('\\', '\\\\')
        keys_list = self.cache.keys(pattern=r'{}'.format(pattern))
        for key in keys_list:
            yield key.decode("UTF-8")

    def get(self, key):
        """
        get the value of the key from the cache
        :param key: str or int type of key
        :return: the value of the key
        """
        res = self.cache.get(r'{}'.format(key))
        if res is not None:
            try:
                return json.loads(res.decode("UTF-8"))
            except:
                return res.decode("UTF-8")

    def delete(self, key):
        """
        delete the element from the cache
        :param key: str or int type of key
        :return:
        """
        self.cache.delete(key)

    def keys(self):
        """
        return all the cache keys
        :return: list of the keys
        """
        for output in list(self.cache.keys()):
            if output is not None:
                yield output.decode('utf-8')

    def clear(self):
        """
        return all the cache keys
        :return: list of the keys
        """
        all_keys = self.cache.keys('*')
        for k in all_keys:
            self.cache.delete(k)



import datetime
import logging
import json
import time

import numpy as np

from .. import entities, exceptions

logger = logging.getLogger(name='dtlpy')


def prepare_dataset(dataset: entities.Dataset,
                    subsets: dict,
                    filters: entities.Filters = None,
                    async_clone: bool = False):
    """
    clones the given dataset and locks it to be readonly

    :param dataset:  `dl.Dataset` to clone
    :param dtlpy.entities.filters.Filters filters: `dl.Filters` to use when cloning - which items to clone
    :param subsets: dictionary to set subsets from the data.
                        key is subset name, value is the filter to set, eg.
                                    {dl.DatasetSubsetType.TEST: dl.Filters(field='dir', values='/test')}
    :param async_clone: `bool` if True (default) - waits until the sync is complete
                                     False - return before the cloned dataset is fully populated
    :return: `dl.Dataset` which is readonly
    """

    project = dataset.project
    now = datetime.datetime.now(datetime.timezone.utc).isoformat(timespec='minutes', sep='T')  # This serves as an id
    today = datetime.datetime.now(datetime.timezone.utc).strftime('%F')

    # CLONE
    clone_name = 'cloned-{ds_name}-{date_str}'.format(ds_name=dataset.name, date_str=today)
    try:
        cloned_dataset = project.datasets.get(clone_name)
        logger.warning("Cloned dataset already exist. Using it...")
        return cloned_dataset
    except exceptions.NotFound:
        pass

    tic = time.time()
    cloned_dataset = dataset.clone(clone_name=clone_name,
                                   filters=filters)
    toc = time.time()

    orig_recipe = dataset.recipes.list()[0]
    cloned_recipe = orig_recipe.clone(shallow=True)
    cloned_dataset.metadata['system']['recipes'] = [cloned_recipe.id]

    logger.info("clone complete: {!r} in {:1.1f}[s]".format(cloned_dataset.name, toc - tic))

    assert cloned_dataset.name is not None, ('unable to get new ds {}'.format(clone_name))
    cloned_dataset.metadata['system']['clone_info'] = {'date': now,
                                                       'originalDatasetId': dataset.id}
    if filters is not None:
        cloned_dataset.metadata['system']['clone_info'].update({'filters': json.dumps(filters.prepare())})
    cloned_dataset.update(system_metadata=True)
    return cloned_dataset



import os
import json
import glob
import shutil

from .base_cache import BaseCache


class FileSystemCache(BaseCache):
    def __init__(self, options=None, size=1000, ttl=1000):
        """
        docs
        """
        if options is None:
            options = {}
        root_dir = options.get('rootDir', os.environ['DEFAULT_CACHE_PATH'])
        if not os.path.isdir(root_dir):
            os.makedirs(root_dir, exist_ok=True)
        self.root_dir = root_dir
        self.ttl = ttl
        self.size = size

    def set(self, key, value):
        """
        set or add a key and value to the cache
        :param key: str or int type of key
        :param value: pickled value
        :return:
        """
        if not isinstance(key, str):
            raise ValueError("key must be string")
        filepath = os.path.join(self.root_dir, key + '.json')
        if not os.path.isdir(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath), exist_ok=True)

        with open(filepath, 'w') as f:
            json.dump(value, f, indent=4)

    def list(self, key):
        """
        get the value of the key from the cache
        :param key: str or int type of key
        :return: the value of the key
        """
        if self.root_dir not in key:
            filepath = os.path.join(self.root_dir, key)
        else:
            filepath = key
        files_list = set(glob.glob(filepath + '.*', recursive=True))
        output_values = []
        for file in files_list:
            with open(file) as file_out:
                output_values.append(json.load(file_out))
        return output_values

    def get(self, key):
        """
        get the value of the key from the cache
        :param key: str or int type of key
        :return: the value of the key
        """
        if self.root_dir not in key:
            filepath = os.path.join(self.root_dir, key)
        else:
            filepath = key

        res = None
        if os.path.isfile(filepath):
            with open(filepath) as file_out:
                res = (json.load(file_out))
        return res

    def delete(self, key):
        """
        delete the element from the cache
        :param key: str or int type of key
        :return:
        """
        if self.root_dir not in key:
            filepath = os.path.join(self.root_dir, os.path.dirname(key))
        else:
            filepath = os.path.dirname(key)
        files_list = set(glob.glob(filepath, recursive=True))
        for file in files_list:
            if os.path.isdir(file):
                try:
                    shutil.rmtree(file)
                except OSError as e:
                    print("Error: %s - %s." % (e.filename, e.strerror))



import numpy as np
# import open3d as o3d
from . import BaseAnnotationDefinition
# from scipy.spatial.transform import Rotation as R
import logging

logger = logging.getLogger(name='dtlpy')


class Cube3d(BaseAnnotationDefinition):
    """
        Cube annotation object
    """
    type = "cube_3d"

    def __init__(self, label, position, scale, rotation,
                 attributes=None, description=None):
        """
        :param label:
        :param position: the XYZ position of the ‘center’ of the annotation.
        :param scale: the scale of the object by each axis (XYZ).
        :param rotation: an euler representation of the object rotation on each axis (with rotation order ‘XYZ’). (rotation in radians)
        :param attributes:
        :param description:
        """
        super().__init__(description=description, attributes=attributes)

        self.position = position
        self.scale = scale
        self.rotation = rotation
        self.label = label

    def _translate(self, points, translate_x, translate_y, translate_z):
        translation_matrix = np.array([[1, 0, 0, 0],
                                       [0, 1, 0, 0],
                                       [0, 0, 1, 0],
                                       [translate_x, translate_y, translate_z, 1]])

        matrix = [(list(i) + [1]) for i in points]
        pts2 = np.dot(matrix, translation_matrix)
        return [pt[:3] for pt in pts2]

    # def make_points(self):
    #     simple = [
    #         [self.scale[0] / 2, self.scale[1] / 2, self.scale[2] / 2],
    #         [-self.scale[0] / 2, self.scale[1] / 2, self.scale[2] / 2],
    #         [self.scale[0] / 2, -self.scale[1] / 2, self.scale[2] / 2],
    #         [self.scale[0] / 2, self.scale[1] / 2, -self.scale[2] / 2],
    #         [-self.scale[0] / 2, -self.scale[1] / 2, self.scale[2] / 2],
    #         [self.scale[0] / 2, -self.scale[1] / 2, -self.scale[2] / 2],
    #         [-self.scale[0] / 2, self.scale[1] / 2, -self.scale[2] / 2],
    #         [-self.scale[0] / 2, -self.scale[1] / 2, -self.scale[2] / 2],
    #     ]
    #
    #     # matrix = R.from_euler('xyz', self.rotation, degrees=False)
    #
    #     vecs = [np.array(p) for p in simple]
    #     rotated = matrix.apply(vecs)
    #     translation = np.array(self.position)
    #     dX = translation[0]
    #     dY = translation[1]
    #     dZ = translation[2]
    #     points = self._translate(rotated, dX, dY, dZ)
    #     return points

    @property
    def geo(self):
        return np.asarray([
            list(self.position),
            list(self.scale),
            list(self.rotation)
        ])

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise
        points = self.make_points()
        front_bl = points[0]
        front_br = points[1]
        front_tr = points[2]
        front_tl = points[3]
        back_bl = points[4]
        back_br = points[5]
        back_tr = points[6]
        back_tl = points[7]
        logger.warning('the show for 3d_cube is not supported.')
        return image

        # image = np.zeros((100, 100, 100), dtype=np.uint8)
        # pcd = o3d.io.read_point_cloud(r"C:\Users\97250\PycharmProjects\tt\qw\3D\D34049418_0000635.las.pcd")
        # # o3d.visualization.draw_geometries([pcd])
        # # points = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1],
        # #           [0, 1, 1], [1, 1, 1]]
        # lines = [[0, 1], [0, 2], [1, 3], [2, 3], [4, 5], [4, 6], [5, 7], [6, 7],
        #          [0, 4], [1, 5], [2, 6], [3, 7]]
        # colors = [[1, 0, 0] for i in range(len(lines))]
        # points = [back_bl, back_br, back_tl, back_tr, front_bl, front_br, front_tl, front_tr]
        # line_set = o3d.geometry.LineSet()
        # line_set.points = o3d.utility.Vector3dVector(points)
        # line_set.lines = o3d.utility.Vector2iVector(lines)
        # line_set.colors = o3d.utility.Vector3dVector(colors)
        # o3d.visualization.draw_geometries([line_set])
        # return image

    def to_coordinates(self, color=None):
        keys = ["position", "scale", "rotation"]
        coordinates = {keys[idx]: {"x": float(x), "y": float(y), "z": float(z)}
                       for idx, [x, y, z] in enumerate(self.geo)}
        return coordinates

    @staticmethod
    def from_coordinates(coordinates):
        geo = list()
        for key, pt in coordinates.items():
            geo.append([pt["x"], pt["y"], pt["z"]])
        return np.asarray(geo)

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            key = "coordinates"
        elif "data" in _json:
            key = "data"
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))

        return cls(
            position=list(_json[key]['position'].values()),
            scale=list(_json[key]['scale'].values()),
            rotation=list(_json[key]['rotation'].values()),
            label=_json["label"],
            attributes=_json.get("attributes", None)
        )



from collections import namedtuple
from enum import Enum
import traceback
import logging
import attr

from .. import repositories, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class PodType(str, Enum):
    SMALL = "small"
    MEDIUM = "medium"
    HIGH = "high"


class CacheAction(str, Enum):
    APPLY = "apply"
    DESTROY = "destroy"


class OrganizationsPlans(str, Enum):
    PREMIUM = "premium"
    FREEMIUM = "freemium"


class MemberOrgRole(str, Enum):
    OWNER = "owner"
    ADMIN = "admin"
    MEMBER = "member"
    WORKER = "worker"


@attr.s()
class Organization(entities.BaseEntity):
    """
    Organization entity
    """

    members = attr.ib(type=list)
    groups = attr.ib(type=list)
    account = attr.ib(type=dict)
    created_at = attr.ib()
    updated_at = attr.ib()
    id = attr.ib(repr=False)
    name = attr.ib(repr=False)
    logo_url = attr.ib(repr=False)
    plan = attr.ib(repr=False)
    owner = attr.ib(repr=False)
    creator = attr.ib(repr=False)

    # api
    _client_api = attr.ib(type=ApiClient, repr=False)

    # repositories
    _repositories = attr.ib(repr=False)

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @property
    def createdBy(self):
        return self.creator

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['organizations', 'projects', 'integrations', 'services', 'settings'])

        r = reps(projects=repositories.Projects(client_api=self._client_api, org=self),
                 organizations=repositories.Organizations(client_api=self._client_api),
                 integrations=repositories.Integrations(client_api=self._client_api, org=self),
                 services=repositories.Services(client_api=self._client_api),
                 settings=repositories.Settings(client_api=self._client_api,
                                                org=self,
                                                resource=self,
                                                resource_type=entities.PlatformEntityType.ORG)
                 )
        return r

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("iam/{}/members".format(self.id))

    @property
    def accounts(self):
        return [self.account]

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    @property
    def settings(self):
        assert isinstance(self._repositories.settings, repositories.Settings)
        return self._repositories.settings

    @property
    def organizations(self):
        assert isinstance(self._repositories.organizations, repositories.Organizations)
        return self._repositories.organizations

    @property
    def integrations(self):
        assert isinstance(self._repositories.integrations, repositories.Integrations)
        return self._repositories.integrations

    @staticmethod
    def _protected_from_json(_json, client_api):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param client_api: ApiClient entity

        :return: update status: bool, Organization entity
        """
        try:
            organization = Organization.from_json(_json=_json,
                                                  client_api=client_api)
            status = True
        except Exception:
            organization = traceback.format_exc()
            status = False
        return status, organization

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Project entity object from a json

        :param bool is_fetched: is Entity fetched from Platform
        :param dict _json: _json response from host
        :param dl.ApiClient client_api: ApiClient entity
        :return: Organization object
        :rtype: dtlpy.entities.organization.Organization
        """
        inst = cls(members=_json.get('members', None),
                   groups=_json.get('groups', None),
                   account=_json.get('account', None),
                   created_at=_json.get('createdAt', None),
                   updated_at=_json.get('updatedAt', None),
                   id=_json.get('id', None),
                   name=_json.get('name', None),
                   logo_url=_json.get('logoUrl', None),
                   plan=_json.get('plan', None),
                   owner=_json.get('owner', None),
                   creator=_json.get('creator', None),
                   client_api=client_api)
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        output_dict = attr.asdict(self,
                                  filter=attr.filters.exclude(attr.fields(Organization)._client_api,
                                                              attr.fields(Organization)._repositories,
                                                              attr.fields(Organization).created_at,
                                                              attr.fields(Organization).updated_at,
                                                              attr.fields(Organization).creator,
                                                              ))
        output_dict['members'] = self.members
        output_dict['groups'] = self.groups
        output_dict['account'] = self.account
        output_dict['accounts'] = self.accounts
        output_dict['createdAt'] = self.created_at
        output_dict['updatedAt'] = self.updated_at
        output_dict['id'] = self.id
        output_dict['name'] = self.name
        output_dict['logo_url'] = self.logo_url
        output_dict['plan'] = self.plan
        output_dict['owner'] = self.owner
        output_dict['creator'] = self.creator

        return output_dict

    def list_groups(self):
        """
        List all organization groups (groups that were created within the organization).

        Prerequisites: You must be an organization "owner" to use this method.

        :return: groups list
        :rtype: list

        """
        return self.organizations.list_groups(organization=self)

    def list_members(self, role: MemberOrgRole = None):
        """
        List all organization members.

        Prerequisites: You must be an organization "owner" to use this method.

        :param str role: MemberOrgRole.ADMIN, MemberOrgRole.OWNER, MemberOrgRole.MEMBER, MemberOrgRole.WORKER
        :return: projects list
        :rtype: list
        """
        return self.organizations.list_members(organization=self, role=role)

    def update(self, plan: str):
        """
        Update Organization.

        Prerequisities: You must be an Organization **superuser** to update an organization.

        :param str plan: OrganizationsPlans.FREEMIUM, OrganizationsPlans.PREMIUM

        :return: organization object
        """
        return self.organizations.update(organization=self, plan=plan)

    def add_member(self, email, role: MemberOrgRole = MemberOrgRole):
        """
        Add members to your organization. Read about members and groups [here](https://dataloop.ai/docs/org-members-groups).

        Prerequisities: To add members to an organization, you must be in the role of an "owner" in that organization.

        :param str email: the member's email
        :param str role: MemberOrgRole.ADMIN, MemberOrgRole.OWNER, MemberOrgRole.MEMBER, MemberOrgRole.WORKER
        :return: True if successful or error if unsuccessful
        :rtype: bool
        """
        return self.organizations.add_member(organization=self, email=email, role=role)

    def delete_member(self, user_id: str, sure: bool = False, really: bool = False):
        """
        Delete member from the Organization.

        Prerequisites: Must be an organization "owner" to delete members.

        :param str user_id: user id
        :param bool sure: Are you sure you want to delete?
        :param bool really: Really really sure?
        :return: True if success and error if not
        :rtype: bool
        """
        return self.organizations.delete_member(organization=self, user_id=user_id, sure=sure, really=really)

    def update_member(self, email: str, role: MemberOrgRole = MemberOrgRole.MEMBER):
        """
        Update member role.

        Prerequisities: You must be an organization "owner" to update a member's role.

        :param str email: the member's email
        :param str role: MemberOrgRole.ADMIN, MemberOrgRole.OWNER, MemberOrgRole.MEMBER, MemberOrgRole.WORKER
        :return: json of the member fields
        :rtype: dict
        """
        return self.organizations.update_member(organization=self, email=email, role=role)

    def open_in_web(self):
        """
        Open the organizations in web platform

        """
        self._client_api._open_in_web(url=self.platform_url)

    def cache_action(self, mode=CacheAction.APPLY, pod_type=PodType.SMALL):
        """
        Open the organizations in web platform

        :param str mode: dl.CacheAction.APPLY or dl.CacheAction.DESTROY
        :param dl.PodType pod_type:  dl.PodType.SMALL, dl.PodType.MEDIUM, dl.PodType.HIGH
        :return: True if success
        :rtype: bool
        """
        return self.services._cache_action(organization=self, mode=mode, pod_type=pod_type)



def main(project_name, dataset_name):
    # Imports the SDK package
    import dtlpy as dl

    """
    Label dictionary format:
    
    {
        'label_name': 'dog',
        'displayLabel': 'Dog',
        'attributes': ['list of attributes'],
        'color': (34, 6, 231),
        'children': ['list of label dictionaries']
    }

    """

    # prep
    project = dl.projects.get(project_name=project_name)
    dataset = project.datasets.get(dataset_name=dataset_name)

    #########################
    # View dataset's labels #
    #########################
    # as objects
    labels = dataset.labels

    # as instance map
    labels = dataset.instance_map

    ###############################
    # add label to dataset entity #
    ###############################
    dataset.add_label(label_name='Horse', color=(2, 43, 123))

    #############################
    # add label with sub-labels #
    #############################
    dataset.add_label(label_name='CEO', color=(2, 43, 123),
                      children=[{'label_name': 'Manager',
                                 'children': [{'label_name': 'Employee'}]}])

    ################################
    # add labels to dataset entity #
    ################################
    labels = [
        {'label_name': 'Dog',
         'color': (34, 6, 231),
         'children': [{'label_name': 'Puppy',
                       'color': (24, 16, 130)}]},
        {'label_name': 'Cat',
         'color': (24, 25, 31),
         'children': [{'label_name': 'Kitten',
                       'color': (124, 116, 140)}]}
    ]
    dataset.add_labels(label_list=labels)

    #################
    # Delete Labels #
    #################
    dataset.delete_labels(label_names=['Cat', 'Dog'])

    ########################################
    # Copy dataset labels to a new dataset #
    ########################################
    new_dataset = project.datasets.create(dataset_name='new_dataset_with_labels',
                                          labels=dataset.labels)

    ##########################################
    # Copy dataset ontology to a new dataset #
    ##########################################
    new_dataset = project.datasets.create(dataset_name='new_dataset_with_ontology',
                                          ontology_ids=dataset.ontology_ids)

    ##############################################
    # Copy dataset labels to an existing dataset #
    ##############################################
    new_dataset = project.datasets.create(dataset_name='new_dataset_without_labels')

    # Get from a list or recipes
    recipe = new_dataset.recipes.list()[0]
    # Or get recipe by id
    # recipe = new_dataset.recipes.get(recipe_id='recipe_id')

    # Get from the list of ontologies
    ontology = recipe.ontologies.list()[0]
    # Or get ontology by id
    # ontology = recipe.ontologies.get(ontology='ontology_id')

    # Add the labels to the dataset
    ontology.add_labels(label_list=dataset.labels)
    ontology.update()

    #########################################################
    # Copy dataset ontology to an existing dataset's recipe #
    #########################################################
    new_dataset = project.datasets.create(dataset_name='new_dataset_without_ontology')
    # get recipe
    new_dataset_recipe = new_dataset.recipes.list()[0]
    # Copy from a different dataset
    new_dataset_recipe.ontology_ids = dataset.ontology_ids
    # Update the new dataset
    new_dataset_recipe.update()

    ##########################
    # update existing recipe #
    ##########################
    # Get recipe from list
    recipe = dataset.recipes.list()[0]
    # Or get specific recipe:
    # recipe = dataset.recipes.get(recipe_id='id')

    # Get ontology from list
    ontology = recipe.ontologies.list()[0]
    # Or get specific ontology:
    # ontology = recipe.ontologies.get(ontology_id='id')

    # Add one label
    ontology.add_label(label_name='Lion', color=(35, 234, 123))

    # Add a list of labels
    labels = [{'label_name': 'Shark', 'color': (1, 1, 1)}, {'label_name': 'Whale', 'color': (34, 56, 7)},
              {'label_name': 'Dolphin', 'color': (100, 14, 150)}]
    ontology.add_labels(label_list=labels)

    # After adding - update ontology
    ontology.update()

    #####################
    # Create new recipe #
    #####################
    # Label list
    labels = [{'tag': 'Donkey', 'color': (1, 1, 1)}, {'tag': 'Mammoth', 'color': (34, 56, 7)},
              {'tag': 'Bird', 'color': (100, 14, 150)}]
    recipe = dataset.recipes.create(recipe_name='My Recipe', labels=labels)



class ReflectDict(dict):

    def __init__(self, value_type: type, start: int = None, end: int = None, on_access: callable = None):
        super(ReflectDict, self).__init__()
        self.value_type = value_type
        self.on_access = on_access
        self._start = int(start) if start is not None else 0
        self._end = int(end) if end is not None else 0

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        if not isinstance(start, float) and not isinstance(start, int):
            raise ValueError('Must input a valid number')
        self._start = int(start) if start is not None else 0

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        if not isinstance(end, float) and not isinstance(end, int):
            raise ValueError('Must input a valid number')
        self._end = int(end) if end is not None else 0

    def actual_keys(self):
        return super(ReflectDict, self).keys()

    def keys(self):
        sorted_keys = list(super(ReflectDict, self).keys())
        sorted_keys.sort()

        yield self._start
        last_yielded = self._start

        for key in sorted_keys:
            if last_yielded == key - 1:
                last_yielded = key
                yield key
            else:
                while last_yielded < key:
                    last_yielded = last_yielded + 1
                    yield last_yielded

        while last_yielded < self._end:
            last_yielded = last_yielded + 1
            yield last_yielded

    def values(self):
        for key in self.keys():
            yield self[key]

    def items(self):
        for key in self.keys():
            yield key, self[key]

    def __iter__(self):
        return self.keys()

    def __contains__(self, key):
        return key in list(self.keys())

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise Exception('Key Error - key must be an integer')

        if key > self._end:
            self._end = key

        if key < self._start:
            self._start = key

        super(ReflectDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        requested_key = key
        if not isinstance(key, int):
            raise Exception('Key Error - key must be an integer')

        if key < self._start or key > self._end:
            raise KeyError(key)
        elif super(ReflectDict, self).__contains__(key):
            return super(ReflectDict, self).__getitem__(key)
        else:
            while key > self._start:
                key = key - 1
                if super(ReflectDict, self).__contains__(key):
                    item = super(ReflectDict, self).__getitem__(key)
                    if isinstance(item, self.value_type):
                        if self.on_access is not None:
                            item = self.on_access(self, actual_key=key, requested_key=requested_key, val=item)
                        return item
                    else:
                        raise Exception('Unknown value type, dict must be of type: {}'.format(
                            self.value_type))

    def __len__(self):
        return len(list(self.keys()))



def main():
    import dtlpy as dl

    project = dl.projects.get(project_name='project_name')

    ##################
    # create trigger #
    ##################
    # create Item trigger
    # that will trigger service with given id
    # when item is created
    trigger = project.triggers.create(service_ids=['some_service_id'],
                                      resource=dl.FiltersResource.ITEM,
                                      actions=dl.TriggerAction.CREATED,
                                      active=True)

    # create Annotation trigger
    # that will trigger service with given id
    # when Annotation is deleted
    trigger = project.triggers.create(service_ids=['some_service_id'],
                                      resource=dl.FiltersResource.ANNOTATION,
                                      actions=dl.TriggerAction.DELETED,
                                      active=True)

    ##################
    # update trigger #
    ##################
    # if we want trigger to be triggered on Created
    trigger.actions = ['Annotation']
    trigger.update()

    # update trigger filters
    # to work only on specific items
    # trigger will be triggered only on
    # items with string: dog in their name
    filters = dl.Filters()
    filters.add(field='filename', values='*dog*')
    trigger.filters = filters
    trigger.update()

    ##################
    # delete trigger #
    ##################
    trigger.delete()

    ############################
    # list of project triggers #
    ############################
    project.triggers.list()



import attr
import logging
import traceback

from .. import repositories, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


@attr.s
class ResourceExecution(entities.BaseEntity):
    """
    Resource execution entity
    """
    # platform
    resource_id = attr.ib()
    resource_type = attr.ib()
    execution_id = attr.ib()
    status = attr.ib()
    timestamp = attr.ib()
    progress = attr.ib()
    message = attr.ib()
    error = attr.ib()
    service_name = attr.ib()
    function_name = attr.ib()
    module_name = attr.ib()
    package_name = attr.ib()
    org_name = attr.ib()
    creator = attr.ib()
    project_id = attr.ib()

    # sdk
    _client_api = attr.ib(type=ApiClient, repr=False)
    _project = attr.ib()
    resource = attr.ib()

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, resource=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param client_api: ApiClient entity
        :return:
        """
        try:
            execution = ResourceExecution.from_json(_json=_json,
                                                    client_api=client_api,
                                                    project=project,
                                                    resource=resource,
                                                    is_fetched=is_fetched)
            status = True
        except Exception:
            execution = traceback.format_exc()
            status = False
        return status, execution

    @classmethod
    def from_json(cls, _json, client_api, project=None, resource=None, is_fetched=True):
        """
        :param dict _json: platform json
        :param dl.ApiClient client_api: ApiClient entity
        :param dtlpy.entities.project.Project project: project entity
        :param entity resource: an entity object (item, dataset, ...)
        :param is_fetched: is Entity fetched from Platform
        """
        if project is not None:
            if project.id != _json.get('projectId', None):
                logger.warning('Execution has been fetched from a project that is not belong to it')
                project = None
        inst = cls(
            resource_id=_json.get('resourceId', None),
            resource_type=_json.get('resourceType', None),
            execution_id=_json.get('executionId', None),
            status=_json.get('status', None),
            timestamp=_json.get('timestamp', None),
            progress=_json.get('progress', None),
            message=_json.get('message', None),
            error=_json.get('error', None),
            service_name=_json.get('serviceName', None),
            function_name=_json.get('functionName', None),
            module_name=_json.get('moduleName', None),
            package_name=_json.get('packageName', None),
            org_name=_json.get('orgName', None),
            creator=_json.get('creator', None),
            project_id=_json.get('projectId', None),
            client_api=client_api,
            project=project,
            resource=resource,
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        # get excluded
        _json = attr.asdict(
            self, filter=attr.filters.exclude(
                attr.fields(ResourceExecution)._client_api,
                attr.fields(ResourceExecution)._project,
                attr.fields(ResourceExecution).resource_id,
                attr.fields(ResourceExecution).resource_type,
                attr.fields(ResourceExecution).execution_id,
                attr.fields(ResourceExecution).function_name,
                attr.fields(ResourceExecution).service_name,
                attr.fields(ResourceExecution).module_name,
                attr.fields(ResourceExecution).package_name,
                attr.fields(ResourceExecution).org_name,
                attr.fields(ResourceExecution).project_id,
            )
        )

        # rename
        _json['projectId'] = self.project_id
        _json['resourceId'] = self.resource_id
        _json['resourceType'] = self.resource_type
        _json['functionName'] = self.function_name
        _json['executionId'] = self.execution_id
        _json['serviceName'] = self.service_name
        _json['moduleName'] = self.module_name
        _json['packageName'] = self.package_name
        _json['orgName'] = self.org_name

        return _json

    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project



import datetime
import logging
import json
from typing import Union, List
import warnings

from .. import exceptions, miscellaneous, entities, repositories, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name="dtlpy")
URL_PATH = "/annotationtasks"


class Tasks:
    """
    Tasks Repository

    The Tasks class allows the user to manage tasks and their properties.
    For more information, read in our developers' documentation about `Creating Tasks <https://developers.dataloop.ai/tutorials/task_workflows/create_a_task/chapter/>`_, and `Redistributing and Reassigning Tasks <https://developers.dataloop.ai/tutorials/task_workflows/redistributing_and_reassigning_a_task/chapter/>`_.
    """

    def __init__(
        self,
        client_api: ApiClient,
        project: entities.Project = None,
        dataset: entities.Dataset = None,
        project_id: str = None,
    ):
        self._client_api = client_api
        self._project = project
        self._dataset = dataset
        self._assignments = None
        self._project_id = project_id

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None and self._project_id is None:
            if self._dataset is None:
                raise exceptions.PlatformException(
                    error="2001",
                    message='Missing "project". need to set a Project entity or use project.tasks repository',
                )
            else:
                self._project = self._dataset.project
                self._project_id = self._project.id
        if self._project is None and self._project_id is not None:
            self._project = self._client_api.projects.get(project_id=self._project_id)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError("Must input a valid Project entity")
        self._project = project

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            raise exceptions.PlatformException(
                error="2001", message='Missing "dataset". need to set a Dataset entity or use dataset.tasks repository'
            )
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        if not isinstance(dataset, entities.Dataset):
            raise ValueError("Must input a valid Dataset entity")
        self._dataset = dataset

    @property
    def assignments(self) -> repositories.Assignments:
        if self._assignments is None:
            self._assignments = repositories.Assignments(client_api=self._client_api, project=self._project)
        assert isinstance(self._assignments, repositories.Assignments)
        return self._assignments

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Task]:
        pool = self._client_api.thread_pools(pool_name="entity.create")
        jobs = [None for _ in range(len(response_items))]

        for i_task, task in enumerate(response_items):
            jobs[i_task] = pool.submit(
                entities.Task._protected_from_json,
                **{"client_api": self._client_api, "_json": task, "project": self._project, "dataset": self._dataset},
            )

        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        tasks = miscellaneous.List([r[1] for r in results if r[0] is True])
        return tasks

    def _list(self, filters: entities.Filters):
        url = "{}/query".format(URL_PATH)
        query = filters.prepare()
        query["context"] = dict(projectIds=[self._project_id])
        success, response = self._client_api.gen_request(req_type="post", path=url, json_req=filters.prepare())

        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def query(self, filters=None, project_ids=None):
        """
        List all tasks by filter.

        **Prerequisites**: You must be in the role of an *owner* or *developer* or *annotation manager* who has been assigned the task.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param list project_ids: list of project ids of the required tasks
        :return: Paged entity - task pages generator
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            dataset.tasks.query(project_ids='project_ids')
        """
        if project_ids is None:
            if self._project_id is not None:
                project_ids = self._project_id
            else:
                project_ids = self.project.id

        if not isinstance(project_ids, list):
            project_ids = [project_ids]

        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.TASK)
        else:
            if not isinstance(filters, entities.Filters):
                raise exceptions.PlatformException("400", "Unknown filters type")
            if filters.resource != entities.FiltersResource.TASK:
                raise exceptions.PlatformException("400", "Filter resource must be task")

        if filters.context is None:
            filters.context = {"projectIds": project_ids}

        if self._project_id is not None:
            filters.add(field="projectId", values=self._project_id)

        if self._dataset is not None:
            filters.add(field="datasetId", values=self._dataset.id)

        paged = entities.PagedEntities(
            items_repository=self,
            filters=filters,
            page_offset=filters.page,
            page_size=filters.page_size,
            project_id=self._project_id,
            client_api=self._client_api,
        )
        paged.get_page()
        return paged

    ###########
    # methods #
    ###########
    @_api_reference.add(path="/annotationtasks/query", method="post")
    def list(
        self,
        project_ids=None,
        status=None,
        task_name=None,
        pages_size=None,
        page_offset=None,
        recipe=None,
        creator=None,
        assignments=None,
        min_date=None,
        max_date=None,
        filters: entities.Filters = None,
    ) -> Union[miscellaneous.List[entities.Task], entities.PagedEntities]:
        """
        List all tasks.

        **Prerequisites**: You must be in the role of an *owner* or *developer* or *annotation manager* who has been assigned the task.

        :param project_ids: search tasks by given list of project ids
        :param str status: search tasks by a given task status
        :param str task_name: search tasks by a given task name
        :param int pages_size: pages size of the output generator
        :param int page_offset: page offset of the output generator
        :param dtlpy.entities.recipe.Recipe recipe: Search tasks that use a given recipe. Provide the required recipe object
        :param str creator: search tasks created by a given creator (user email)
        :param dtlpy.entities.assignment.Assignment recipe assignments: assignments object
        :param double min_date: search all tasks created AFTER a given date, use a milliseconds format. For example: 1661780622008
        :param double max_date: search all tasks created BEFORE a given date, use a milliseconds format. For example: 1661780622008
        :param dtlpy.entities.filters.Filters filters: dl.Filters entity to filters tasks using DQL
        :return: List of Task objects

        **Example**:

        .. code-block:: python

            dataset.tasks.list(project_ids='project_ids',pages_size=100, page_offset=0)
        """
        # url
        url = URL_PATH + "/query"

        if filters is None:
            filters = entities.Filters(use_defaults=False, resource=entities.FiltersResource.TASK)
        else:
            return self.query(filters=filters, project_ids=project_ids)

        if self._dataset is not None:
            filters.add(field="datasetId", values=self._dataset.id)

        if project_ids is not None:
            if not isinstance(project_ids, list):
                project_ids = [project_ids]
        elif self._project_id is not None:
            project_ids = [self._project_id]
        else:
            project_ids = [self.project.id]
        filters.context = {"projectIds": project_ids}

        if assignments is not None:
            if not isinstance(assignments, list):
                assignments = [assignments]
            assignments = [
                assignments_entity.id if isinstance(assignments_entity, entities.Assignment) else assignments_entity
                for assignments_entity in assignments
            ]
            filters.add(field="assignmentIds", values=assignments, operator=entities.FiltersOperations.IN)
        if status is not None:
            filters.add(field="status", values=status)
        if task_name is not None:
            filters.add(field="name", values=task_name)
        if pages_size is not None:
            filters.page_size = pages_size
        if pages_size is None:
            filters.page_size = 500
        if page_offset is not None:
            filters.page = page_offset
        if recipe is not None:
            if not isinstance(recipe, list):
                recipe = [recipe]
            recipe = [
                recipe_entity.id if isinstance(recipe_entity, entities.Recipe) else recipe_entity
                for recipe_entity in recipe
            ]
            filters.add(field="recipeId", values=recipe, operator=entities.FiltersOperations.IN)
        if creator is not None:
            filters.add(field="creator", values=creator)
        if min_date is not None:
            filters.add(field="dueDate", values=min_date, operator=entities.FiltersOperations.GREATER_THAN)
        if max_date is not None:
            filters.add(field="dueDate", values=max_date, operator=entities.FiltersOperations.LESS_THAN)

        success, response = self._client_api.gen_request(req_type="post", path=url, json_req=filters.prepare())
        if success:
            tasks = miscellaneous.List(
                [
                    entities.Task.from_json(
                        client_api=self._client_api, _json=_json, project=self._project, dataset=self._dataset
                    )
                    for _json in response.json()["items"]
                ]
            )
        else:
            logger.error("Platform error getting annotation task")
            raise exceptions.PlatformException(response)

        return tasks

    @_api_reference.add(path="/annotationtasks/{id}", method="get")
    def get(self, task_name=None, task_id=None) -> entities.Task:
        """
        Get a Task object to use in your code.

        **Prerequisites**: You must be in the role of an *owner* or *developer* or *annotation manager* who has been assigned the task.

        :param str task_name: optional - search by name
        :param str task_id: optional - search by id
        :return: task object
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            dataset.tasks.get(task_id='task_id')
        """

        # url
        url = URL_PATH

        if task_id is not None:
            url = "{}/{}".format(url, task_id)
            success, response = self._client_api.gen_request(req_type="get", path=url)
            if not success:
                raise exceptions.PlatformException(response)
            else:
                task = entities.Task.from_json(
                    _json=response.json(), client_api=self._client_api, project=self._project, dataset=self._dataset
                )
            # verify input task name is same as the given id
            if task_name is not None and task.name != task_name:
                logger.warning(
                    "Mismatch found in tasks.get: task_name is different then task.name:"
                    " {!r} != {!r}".format(task_name, task.name)
                )
        elif task_name is not None:
            tasks = self.list(
                filters=entities.Filters(field="name", values=task_name, resource=entities.FiltersResource.TASK)
            )
            if tasks.items_count == 0:
                raise exceptions.PlatformException("404", "Annotation task not found")
            elif tasks.items_count > 1:
                raise exceptions.PlatformException(
                    "404", f"More than one Annotation task exist with the same name: {task_name}"
                )
            else:
                task = tasks[0][0]
        else:
            raise exceptions.PlatformException("400", "Must provide either Annotation task name or Annotation task id")

        assert isinstance(task, entities.Task)
        return task

    @property
    def platform_url(self):
        return self._client_api._get_resource_url(f"projects/{self.project.id}/tasks")

    def open_in_web(self, task_name: str = None, task_id: str = None, task: entities.Task = None):
        """
        Open the task in the web platform.

        **Prerequisites**: You must be in the role of an *owner* or *developer* or *annotation manager* who has been assigned the task.

        :param str task_name: the name of the task
        :param str task_id: the Id of the task
        :param dtlpy.entities.task.Task task: the task object

        **Example**:

        .. code-block:: python

            dataset.tasks.open_in_web(task_id='task_id')
        """
        if task_name is not None:
            task = self.get(task_name=task_name)
        if task is not None:
            task.open_in_web()
        elif task_id is not None:
            self._client_api._open_in_web(url=self.platform_url + "/" + str(task_id))
        else:
            self._client_api._open_in_web(url=self.platform_url)

    @_api_reference.add(path="/annotationtasks/{id}", method="delete")
    def delete(self, task: entities.Task = None, task_name: str = None, task_id: str = None, wait: bool = True):
        """
        Delete the Task.

        **Prerequisites**: You must be in the role of an *owner* or *developer* or *annotation manager* who created that task.

        :param dtlpy.entities.task.Task task: the task object
        :param str task_name: the name of the task
        :param str task_id: the Id of the task
        :param bool wait: wait until delete task finish
        :return: True is success
        :rtype: bool

        **Example**:

        .. code-block:: python

            dataset.tasks.delete(task_id='task_id')
        """
        if task_id is None:
            if task is None:
                if task_name is None:
                    raise exceptions.PlatformException(
                        "400", "Must provide either annotation task, " "annotation task name or annotation task id"
                    )
                else:
                    task = self.get(task_name=task_name)
            task_id = task.id

        url = URL_PATH
        url = f"{url}/{task_id}"
        success, response = self._client_api.gen_request(req_type="delete", path=url, json_req={"asynced": wait})

        if not success:
            raise exceptions.PlatformException(response)
        response_json = response.json()
        command = entities.Command.from_json(_json=response_json, client_api=self._client_api)
        if not wait:
            return command
        command = command.wait(timeout=0)
        if "deleteTaskId" not in command.spec:
            raise exceptions.PlatformException(
                error="400", message="deleteTaskId key is missing in command response: {}".format(response)
            )
        return True

    @_api_reference.add(path="/annotationtasks/{id}", method="patch")
    def update(self, task: entities.Task = None, system_metadata=False) -> entities.Task:
        """
        Update a Task.

        **Prerequisites**: You must be in the role of an *owner* or *developer* or *annotation manager* who created that task.

        :param dtlpy.entities.task.Task task: the task object
        :param bool system_metadata: DEPRECATED
        :return: Task object
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            dataset.tasks.update(task='task_entity')
        """
        url = URL_PATH
        url = f"{url}/{task.id}"

        if system_metadata:
            warnings.warn(
                "Task system metadata updates are not permitted. Please store custom metadata in 'task.metadata['user']' instead.",
                DeprecationWarning,
            )

        success, response = self._client_api.gen_request(req_type="patch", path=url, json_req=task.to_json())
        if success:
            return entities.Task.from_json(
                _json=response.json(), client_api=self._client_api, project=self._project, dataset=self._dataset
            )
        else:
            raise exceptions.PlatformException(response)

    def create_qa_task(self,
                       task: entities.Task,
                       assignee_ids,
                       due_date=None,
                       filters=None,
                       items=None,
                       query=None,
                       workload=None,
                       metadata=None,
                       available_actions=None,
                       wait=True,
                       batch_size=None,
                       max_batch_workload=None,
                       allowed_assignees=None,
                       priority=entities.TaskPriority.MEDIUM
                       ) -> entities.Task:
        """
        Create a new QA Task.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param dtlpy.entities.task.Task task: the parent annotation task object
        :param list assignee_ids: list the QA task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param float due_date: date by which the QA task should be finished; for example, due_date=datetime.datetime(day=1, month=1, year=2029).timestamp()
        :param entities.Filters filters: dl.Filters entity to filter items for the task
        :param List[entities.Item] items: list of items (item Id or objects) to insert to the task
        :param dict DQL query: filter items for the task
        :param List[WorkloadUnit] workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees. For example: [dl.WorkloadUnit(annotator@hi.com, 80), dl.WorkloadUnit(annotator2@hi.com, 20)]
        :param dict metadata: metadata for the task
        :param list available_actions: list of available actions (statuses) that will be available for the task items; The default statuses are: "approved" and "discard"
        :param bool wait: wait until create task finish
        :param int batch_size: Pulling batch size (items), use with pulling allocation method. Restrictions - Min 3, max 100
        :param int max_batch_workload: Max items in assignment, use with pulling allocation method. Restrictions - Min batchSize + 2, max batchSize * 2
        :param list allowed_assignees: list the task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param entities.TaskPriority priority: priority of the task options in entities.TaskPriority
        :return: task object
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            dataset.tasks.create_qa_task(task= 'task_entity',
                                        due_date = datetime.datetime(day= 1, month= 1, year= 2029).timestamp(),
                                        assignee_ids =[ 'annotator1@dataloop.ai', 'annotator2@dataloop.ai'])
        """
        source_filter = entities.filters.SingleFilter(
            field='metadata.system.refs',
            values={
                "id": task.id,
                "type": "task",
                "metadata":
                    {
                        "status":
                            {
                                "$exists": True
                            }
                    }
            },
            operator=entities.FiltersOperations.MATCH
        )

        if query is not None:
            and_list = query.get('filter', query).get('$and', None)
            if and_list is not None:
                and_list.append(source_filter.prepare())
            else:
                if 'filter' not in query:
                    query['filter'] = {}
                query['filter']['$and'] = [source_filter.prepare()]

        else:
            if filters is None and items is None:
                filters = entities.Filters()
            if filters:
                filters.and_filter_list.append(source_filter)

        return self.create(task_name='{}_qa'.format(task.name),
                           task_type='qa',
                           task_parent_id=task.id,
                           assignee_ids=assignee_ids,
                           workload=workload,
                           task_owner=task.creator,
                           project_id=task.project_id,
                           recipe_id=task.recipe_id,
                           due_date=due_date,
                           filters=filters,
                           items=items,
                           query=query,
                           metadata=metadata,
                           available_actions=available_actions,
                           wait=wait,
                           batch_size=batch_size,
                           max_batch_workload=max_batch_workload,
                           allowed_assignees=allowed_assignees,
                           priority=priority
                           )

    def create_honeypot_task(
        self,
        name: str,
        dataset: entities.Dataset = None,
        due_date: float = None,
        filters: entities.Filters = None,
        owner: str = None,
        recipe_id: str = None,
        assignee_ids: List[str] = None,
        workload=None,
        available_actions=None,
        priority=entities.TaskPriority.MEDIUM,
        consensus_percentage=None,
        consensus_assignees=None,
        scoring=True,
        limit=None,
        wait=True,
        enforce_video_conversion=True,
    ) -> entities.Task:
        """
        Create a new Consensus Task.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param str name: the name of the task
        :param entities.Dataset dataset: dataset object, the dataset that refer to the task
        :param float due_date: date by which the task should be finished; for example, due_date=datetime.datetime(day=1, month=1, year=2029).timestamp()
        :param entities.Filters filters: dl.Filters entity to filter items for the task
        :param str owner: task owner. Provide user email
        :param str recipe_id: recipe id for the task
        :param list assignee_ids: list the task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees
        :param list available_actions: list of available actions (statuses) that will be available for the task items
        :param entities.TaskPriority priority: priority of the task options in entities.TaskPriority
        :param str consensus_task_type: consensus task type - "consensus", "qualification", or "honeypot"
        :param int consensus_percentage: percentage of items to be copied to multiple annotators (consensus items)
        :param int consensus_assignees: the number of different annotators per item (number of copies per item)
        :param bool scoring: create a scoring app in project
        :param int limit: the limit items that the task can include
        :param bool wait: wait until create task finish
        :param bool enforce_video_conversion: Enforce WEBM conversion on video items for frame-accurate annotations
        :return: Task object
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            # Create a consensus task
            dataset.tasks.create_consensus_task(name='my_consensus_task',
                                              assignee_ids=['annotator1@dataloop.ai', 'annotator2@dataloop.ai'],
                                              consensus_percentage=66,
                                              consensus_assignees=2)
        """
        return self.create_consensus_task(
            name=name,
            dataset=dataset,
            due_date=due_date,
            filters=filters,
            owner=owner,
            recipe_id=recipe_id,
            assignee_ids=assignee_ids,
            workload=workload,
            available_actions=available_actions,
            priority=priority,
            consensus_task_type=entities.ConsensusTaskType.HONEYPOT,
            consensus_percentage=consensus_percentage,
            consensus_assignees=consensus_assignees,
            scoring=scoring,
            limit=limit,
            wait=wait,
            enforce_video_conversion=enforce_video_conversion,
        )

    def create_qualification_task(
        self,
        name: str,
        dataset: entities.Dataset = None,
        due_date: float = None,
        filters: entities.Filters = None,
        owner: str = None,
        recipe_id: str = None,
        assignee_ids: List[str] = None,
        workload=None,
        available_actions=None,
        priority=entities.TaskPriority.MEDIUM,
        consensus_percentage=None,
        consensus_assignees=None,
        limit=None,
        wait=True,
        enforce_video_conversion=True,
    ) -> entities.Task:
        """
        Create a new Qualification Task.
        """
        return self.create_consensus_task(
            name=name,
            dataset=dataset,
            due_date=due_date,
            filters=filters,
            owner=owner,
            recipe_id=recipe_id,
            assignee_ids=assignee_ids,
            workload=workload,
            available_actions=available_actions,
            priority=priority,
            consensus_task_type=entities.ConsensusTaskType.QUALIFICATION,
            consensus_percentage=consensus_percentage,
            consensus_assignees=consensus_assignees,
            scoring=True,
            limit=limit,
            wait=wait,
            enforce_video_conversion=enforce_video_conversion,
        )

    def create_consensus_task(
        self,
        name: str,
        dataset: entities.Dataset = None,
        due_date: float = None,
        filters: entities.Filters = None,
        owner: str = None,
        recipe_id: str = None,
        assignee_ids: List[str] = None,
        workload=None,
        available_actions=None,
        priority=entities.TaskPriority.MEDIUM,
        metadata=None,
        consensus_task_type: entities.ConsensusTaskType = entities.ConsensusTaskType.CONSENSUS,
        consensus_percentage=None,
        consensus_assignees=None,
        scoring=True,
        limit=None,
        wait=True,
        enforce_video_conversion=True,
    ) -> entities.Task:
        """
        Create a new Consensus Task.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param str name: the name of the task
        :param entities.Dataset dataset: dataset object, the dataset that refer to the task
        :param float due_date: date by which the task should be finished; for example, due_date=datetime.datetime(day=1, month=1, year=2029).timestamp()
        :param entities.Filters filters: dl.Filters entity to filter items for the task
        :param str owner: task owner. Provide user email
        :param str recipe_id: recipe id for the task
        :param list assignee_ids: list the task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees
        :param list available_actions: list of available actions (statuses) that will be available for the task items
        :param entities.TaskPriority priority: priority of the task options in entities.TaskPriority
        :param dict metadata: metadata for the task
        :param str consensus_task_type: consensus task type - "consensus", "qualification", or "honeypot"
        :param int consensus_percentage: percentage of items to be copied to multiple annotators (consensus items)
        :param int consensus_assignees: the number of different annotators per item (number of copies per item)
        :param bool scoring: create a scoring app in project
        :param int limit: the limit items that the task can include
        :param bool wait: wait until create task finish
        :param bool enforce_video_conversion: Enforce WEBM conversion on video items for frame-accurate annotations
        :return: Task object
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            # Create a consensus task
            dataset.tasks.create_consensus_task(name='my_consensus_task',
                                              assignee_ids=['annotator1@dataloop.ai', 'annotator2@dataloop.ai'],
                                              consensus_percentage=66,
                                              consensus_assignees=2)
        """

        if dataset is None:
            dataset = self.dataset

        if due_date is None:
            due_date = (datetime.datetime.now() + datetime.timedelta(days=7)).timestamp()

        if filters is None:
            filters = entities.Filters()

        if owner is None:
            owner = self._client_api.info()["user_email"]

        if recipe_id is None:
            recipe_id = dataset.get_recipe_ids()[0]

        if workload is None and assignee_ids is not None:
            workload = entities.Workload.generate(assignee_ids=assignee_ids)

        # Handle metadata for consensus tasks
        if metadata is None:
            metadata = {}
        if "system" not in metadata:
            metadata["system"] = {}
        if assignee_ids is not None:
            metadata["system"]["allowedAssignees"] = assignee_ids
        if consensus_task_type is not None:
            metadata["system"]["consensusTaskType"] = consensus_task_type
        metadata = self._add_task_metadata_params(
            metadata=metadata, input_value=consensus_percentage, input_name="consensusPercentage"
        )
        metadata = self._add_task_metadata_params(
            metadata=metadata, input_value=consensus_assignees, input_name="consensusAssignees"
        )
        metadata = self._add_task_metadata_params(metadata=metadata, input_value=scoring, input_name="scoring")

        # Create payload for consensus task
        payload = {
            "name": name,
            "query": "{}".format(json.dumps(filters.prepare()).replace("'", '"')),
            "taskOwner": owner,
            "spec": {"type": "annotation"},
            "datasetId": dataset.id,
            "projectId": self.project.id,
            "assignmentIds": [],
            "recipeId": recipe_id,
            "dueDate": due_date * 1000,
            "asynced": wait,
            "priority": priority,
            "percentage": True,
        }

        # Add workload if provided
        if workload:
            payload["workload"] = workload.to_json()

        # Add limit if provided
        if limit:
            payload["limit"] = limit

        # Add available actions if provided
        if available_actions is not None:
            payload["availableActions"] = [action.to_json() for action in available_actions]

        # Handle video conversion
        if not enforce_video_conversion:
            payload["disableWebm"] = not enforce_video_conversion

        # Handle metadata for consensus tasks
        if metadata is not None:
            payload["metadata"] = metadata

        return self._create_task(payload, wait=wait)

    def _add_task_metadata_params(self, metadata, input_value, input_name):
        if input_value is not None and not isinstance(input_value, int):
            raise exceptions.PlatformException(error="400", message=f"{input_name} must be a numbers")
        if input_value is not None:
            metadata["system"][input_name] = input_value
        return metadata

    def create_labeling_task(
        self,
        name: str,
        dataset: entities.Dataset = None,
        due_date: float = None,
        filters: entities.Filters = None,
        owner: str = None,
        recipe_id: str = None,
        assignee_ids: List[str] = None,
        workload=None,
        available_actions=None,
        priority=entities.TaskPriority.MEDIUM,
        metadata=None,
        batch_size=None,
        max_batch_workload=None,
        allowed_assignees=None,
        limit=None,
        wait=True,
        enforce_video_conversion=True,
    ) -> entities.Task:
        """
        Create a new Annotation Task (Distribution or Pulling).

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param str name: the name of the task
        :param entities.Dataset dataset: dataset object, the dataset that refer to the task
        :param float due_date: date by which the task should be finished; for example, due_date=datetime.datetime(day=1, month=1, year=2029).timestamp()
        :param entities.Filters filters: dl.Filters entity to filter items for the task
        :param str owner: task owner. Provide user email
        :param str recipe_id: recipe id for the task
        :param list assignee_ids: list the task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees
        :param list available_actions: list of available actions (statuses) that will be available for the task items
        :param entities.TaskPriority priority: priority of the task options in entities.TaskPriority
        :param dict metadata: metadata for the task
        :param int batch_size: Pulling batch size (items), use with pulling allocation method. Restrictions - Min 3, max 100
        :param int max_batch_workload: Max items in assignment, use with pulling allocation method. Restrictions - Min batchSize + 2, max batchSize * 2
        :param list allowed_assignees: list the task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param int limit: the limit items that the task can include
        :param bool wait: wait until create task finish
        :param bool enforce_video_conversion: Enforce WEBM conversion on video items for frame-accurate annotations
        :return: Task object
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            # Create a distribution task
            dataset.tasks.create_labeling(name='my_distribution_task',
                                        assignee_ids=['annotator1@dataloop.ai', 'annotator2@dataloop.ai'])

            # Create a pulling task
            dataset.tasks.create_labeling(name='my_pulling_task',
                                        assignee_ids=['annotator1@dataloop.ai', 'annotator2@dataloop.ai'],
                                        batch_size=5,
                                        max_batch_workload=7)
        """

        if dataset is None:
            dataset = self.dataset

        if due_date is None:
            due_date = (datetime.datetime.now() + datetime.timedelta(days=7)).timestamp()

        if filters is None:
            filters = entities.Filters()

        if owner is None:
            owner = self._client_api.info()["user_email"]

        if recipe_id is None:
            recipe_id = dataset.get_recipe_ids()[0]

        if workload is None and assignee_ids is not None:
            workload = entities.Workload.generate(assignee_ids=assignee_ids)

        if metadata is None:
            metadata = {}
        if any([batch_size, max_batch_workload]):
            if "system" not in metadata:
                metadata["system"] = {}
            if allowed_assignees is not None or assignee_ids is not None:
                metadata["system"]["allowedAssignees"] = allowed_assignees if allowed_assignees else assignee_ids
            metadata = self._add_task_metadata_params(metadata=metadata, input_value=batch_size, input_name="batchSize")
            metadata = self._add_task_metadata_params(
                metadata=metadata, input_value=max_batch_workload, input_name="maxBatchWorkload"
            )

        # Create payload for annotation task
        payload = {
            "name": name,
            "query": "{}".format(json.dumps(filters.prepare()).replace("'", '"')),
            "taskOwner": owner,
            "spec": {"type": "annotation"},
            "datasetId": dataset.id,
            "projectId": self.project.id,
            "assignmentIds": [],
            "recipeId": recipe_id,
            "dueDate": due_date * 1000,
            "asynced": wait,
            "priority": priority,
        }

        # Add workload if provided
        if workload:
            payload["workload"] = workload.to_json()

        # Add limit if provided
        if limit:
            payload["limit"] = limit

        # Add available actions if provided
        if available_actions is not None:
            payload["availableActions"] = [action.to_json() for action in available_actions]

        # Handle video conversion
        if not enforce_video_conversion:
            payload["disableWebm"] = not enforce_video_conversion

        # Handle metadata for pulling tasks
        if metadata is not None:
            payload["metadata"] = metadata

        return self._create_task(payload, wait=wait)

    def _create_task(self, payload: dict, wait: bool = True) -> entities.Task:
        """
        Private function to create a task from a prepared payload.

        :param dict payload: the prepared payload for task creation
        :param bool wait: whether to wait for task creation to complete
        :return: created Task object
        :rtype: dtlpy.entities.task.Task
        """
        success, response = self._client_api.gen_request(req_type="post", path=URL_PATH, json_req=payload)
        if success:
            response_json = response.json()
            if payload.get("checkIfExist") is not None and "name" in response_json:
                return entities.Task.from_json(
                    _json=response.json(), client_api=self._client_api, project=self._project, dataset=self._dataset
                )

            command = entities.Command.from_json(_json=response_json, client_api=self._client_api)
            if not wait:
                return command
            command = command.wait(timeout=0)
            if "createTaskPayload" not in command.spec:
                raise exceptions.PlatformException(
                    error="400", message="createTaskPayload key is missing in command response: {}".format(response)
                )
            task = self.get(task_id=command.spec["createdTaskId"])
        else:
            raise exceptions.PlatformException(response)

        assert isinstance(task, entities.Task)
        return task

    @_api_reference.add(path="/annotationtasks", method="post")
    def create(
        self,
        task_name,
        due_date=None,
        assignee_ids=None,
        workload=None,
        dataset=None,
        task_owner=None,
        task_type="annotation",
        task_parent_id=None,
        project_id=None,
        recipe_id=None,
        assignments_ids=None,
        metadata=None,
        filters=None,
        items=None,
        query=None,
        available_actions=None,
        wait=True,
        check_if_exist: entities.Filters = False,
        limit=None,
        batch_size=None,
        max_batch_workload=None,
        allowed_assignees=None,
        priority=entities.TaskPriority.MEDIUM,
        consensus_task_type=None,
        consensus_percentage=None,
        consensus_assignees=None,
        scoring=True,
        enforce_video_conversion=True,
    ) -> entities.Task:
        """
        Create a new Task (Annotation or QA).

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param str task_name: the name of the task
        :param float due_date: date by which the task should be finished; for example, due_date=datetime.datetime(day=1, month=1, year=2029).timestamp()
        :param list assignee_ids: list the task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param List[WorkloadUnit] List[WorkloadUnit] workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees. For example: [dl.WorkloadUnit(annotator@hi.com, 80), dl.WorkloadUnit(annotator2@hi.com, 20)]
        :param entities.Dataset dataset: dataset object, the dataset that refer to the task
        :param str task_owner: task owner. Provide user email
        :param str task_type: task type "annotation" or "qa"
        :param str task_parent_id: optional if type is qa - parent annotation task id
        :param str project_id: the Id of the project where task will be created
        :param str recipe_id: recipe id for the task
        :param list assignments_ids: assignments ids to the task
        :param dict metadata: metadata for the task
        :param entities.Filters filters: dl.Filters entity to filter items for the task
        :param List[entities.Item] items: list of items (item Id or objects) to insert to the task
        :param dict DQL query: filter items for the task
        :param list available_actions: list of available actions (statuses) that will be available for the task items; The default statuses are: "completed" and "discard"
        :param bool wait: wait until create task finish
        :param entities.Filters check_if_exist: dl.Filters check if task exist according to filter
        :param int limit: the limit items that the task can include
        :param int  batch_size: Pulling batch size (items), use with pulling allocation method. Restrictions - Min 3, max 100
        :param int max_batch_workload: max_batch_workload: Max items in assignment, use with pulling allocation method. Restrictions - Min batchSize + 2, max batchSize * 2
        :param list allowed_assignees: list the task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param entities.TaskPriority priority: priority of the task options in entities.TaskPriority
        :param entities.ConsensusTaskType consensus_task_type: consensus_task_type of the task options in entities.ConsensusTaskType
        :param int consensus_percentage: percentage of items to be copied to multiple annotators (consensus items)
        :param int consensus_assignees: the number of different annotators per item (number of copies per item)
        :param bool scoring: create a scoring app in project
        :param bool enforce_video_conversion: Enforce WEBM conversion on video items for frame-accurate annotations. WEBM Conversion will be executed as a project service and incurs compute costs. Service compute resources can be set according to planned workload.
        :return: Task object
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            dataset.tasks.create(task= 'task_entity',
                                due_date = datetime.datetime(day= 1, month= 1, year= 2029).timestamp(),
                                assignee_ids =[ 'annotator1@dataloop.ai', 'annotator2@dataloop.ai'],
                                available_actions=[dl.ItemAction("discard"), dl.ItemAction("to-check")])
        """

        if dataset is None and self._dataset is None:
            raise exceptions.PlatformException("400", "Please provide param dataset")
        if due_date is None:
            due_date = (datetime.datetime.now() + datetime.timedelta(days=7)).timestamp()
        if query is None:
            if filters is None and items is None:
                query = entities.Filters().prepare()
            elif filters is None:
                item_list = list()
                if isinstance(items, entities.PagedEntities):
                    for page in items:
                        for item in page:
                            item_list.append(item)
                elif isinstance(items, list):
                    item_list = items
                elif isinstance(items, entities.Item):
                    item_list.append(items)
                else:
                    raise exceptions.PlatformException("400", "Unknown items type")
                query = entities.Filters(
                    field="id",
                    values=[item.id for item in item_list],
                    operator=entities.FiltersOperations.IN,
                    use_defaults=False,
                ).prepare()
            else:
                query = filters.prepare()

        if dataset is None:
            dataset = self._dataset

        if task_owner is None:
            task_owner = self._client_api.info()["user_email"]

        if task_type not in ["annotation", "qa"]:
            raise ValueError('task_type must be one of: "annotation", "qa". got: {}'.format(task_type))

        if recipe_id is None:
            recipe_id = dataset.get_recipe_ids()[0]

        if project_id is None:
            if self._project_id is not None:
                project_id = self._project_id
            else:
                project_id = self.project.id

        if workload is None and assignee_ids is not None:
            workload = entities.Workload.generate(assignee_ids=assignee_ids)

        if assignments_ids is None:
            assignments_ids = list()

        payload = {
            "name": task_name,
            "query": "{}".format(json.dumps(query).replace("'", '"')),
            "taskOwner": task_owner,
            "spec": {"type": task_type},
            "datasetId": dataset.id,
            "projectId": project_id,
            "assignmentIds": assignments_ids,
            "recipeId": recipe_id,
            "dueDate": due_date * 1000,
            "asynced": wait,
            "priority": priority,
        }

        if check_if_exist:
            if check_if_exist.resource != entities.FiltersResource.TASK:
                raise exceptions.PlatformException(
                    "407",
                    "Filter resource for check_if_exist param must be {}, got {}".format(
                        entities.FiltersResource.TASK, check_if_exist.resource
                    ),
                )
            payload["checkIfExist"] = {"query": check_if_exist.prepare()}

        if workload:
            payload["workload"] = workload.to_json()

        if limit:
            payload["limit"] = limit

        if available_actions is not None:
            payload["availableActions"] = [action.to_json() for action in available_actions]

        if task_parent_id is not None:
            payload["spec"]["parentTaskId"] = task_parent_id

        if not enforce_video_conversion:
            payload["disableWebm"] = not enforce_video_conversion

        is_pulling = any([batch_size, max_batch_workload])
        is_consensus = any([consensus_percentage, consensus_assignees, consensus_task_type])
        if is_pulling and is_consensus:
            raise exceptions.PlatformException(error="400", message="Consensus can not work as a pulling task")
        if any([is_pulling, is_consensus]):
            if metadata is None:
                metadata = {}
            if "system" not in metadata:
                metadata["system"] = {}
            if allowed_assignees is not None or assignee_ids is not None:
                metadata["system"]["allowedAssignees"] = allowed_assignees if allowed_assignees else assignee_ids
            if consensus_task_type is not None:
                metadata["system"]["consensusTaskType"] = consensus_task_type
            metadata = self._add_task_metadata_params(metadata=metadata, input_value=batch_size, input_name="batchSize")
            metadata = self._add_task_metadata_params(
                metadata=metadata, input_value=max_batch_workload, input_name="maxBatchWorkload"
            )
            metadata = self._add_task_metadata_params(
                metadata=metadata, input_value=consensus_percentage, input_name="consensusPercentage"
            )
            metadata = self._add_task_metadata_params(
                metadata=metadata, input_value=consensus_assignees, input_name="consensusAssignees"
            )
            metadata = self._add_task_metadata_params(metadata=metadata, input_value=scoring, input_name="scoring")

        if metadata is not None:
            payload["metadata"] = metadata

        return self._create_task(payload, wait=wait)

    def __item_operations(self, dataset: entities.Dataset, op, task=None, task_id=None, filters=None, items=None):

        if task is None and task_id is None:
            raise exceptions.PlatformException("400", "Must provide either task or task id")
        elif task_id is None:
            task_id = task.id

        try:
            if filters is None and items is None:
                raise exceptions.PlatformException("400", "Must provide either filters or items list")

            if filters is None:
                filters = entities.Filters(
                    field="id",
                    values=[item.id for item in items],
                    operator=entities.FiltersOperations.IN,
                    use_defaults=False,
                )

            if op == "delete":
                if task is None:
                    task = self.get(task_id=task_id)
                assignment_ids = task.assignmentIds
                filters._ref_assignment = True
                filters._ref_assignment_id = assignment_ids

            filters._ref_task = True
            filters._ref_task_id = task_id
            filters._ref_op = op
            return dataset.items.update(filters=filters)
        finally:
            if filters is not None:
                filters._nullify_refs()

    @_api_reference.add(path="/annotationtasks/{id}/addToTask", method="post")
    def add_items(
        self,
        task: entities.Task = None,
        task_id=None,
        filters: entities.Filters = None,
        items=None,
        assignee_ids=None,
        query=None,
        workload=None,
        limit=None,
        wait=True,
    ) -> entities.Task:
        """
        Add items to a Task.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param dtlpy.entities.task.Task task: task object
        :param str task_id: the Id of the task
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param list items: list of items (item Ids or objects) to add to the task
        :param list assignee_ids: list to assignee who works in the task
        :param dict query: query to filter the items for the task
        :param list workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees. For example: [dl.WorkloadUnit(annotator@hi.com, 80), dl.WorkloadUnit(annotator2@hi.com, 20)]
        :param int limit: the limit items that task can include
        :param bool wait: wait until add items will to finish
        :return: task entity
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            dataset.tasks.add_items(task= 'task_entity',
                                items = [items])
        """
        if filters is None and items is None and query is None:
            raise exceptions.PlatformException("400", "Must provide either filters, query or items list")

        if task is None and task_id is None:
            raise exceptions.PlatformException("400", "Must provide either task or task_id")

        if query is None:
            if filters is None:
                if not isinstance(items, list):
                    items = [items]
                filters = entities.Filters(
                    field="id",
                    values=[item.id for item in items],
                    operator=entities.FiltersOperations.IN,
                    use_defaults=False,
                )
            query = filters.prepare()

        if workload is None and assignee_ids is not None:
            workload = entities.Workload.generate(assignee_ids=assignee_ids)

        if task_id is None:
            task_id = task.id

        payload = {"query": "{}".format(json.dumps(query).replace("'", '"'))}

        if workload is not None:
            payload["workload"] = workload.to_json()

        if limit is not None:
            payload["limit"] = limit

        payload["asynced"] = wait

        url = "{}/{}/addToTask".format(URL_PATH, task_id)

        success, response = self._client_api.gen_request(req_type="post", path=url, json_req=payload)

        if success:
            command = entities.Command.from_json(_json=response.json(), client_api=self._client_api)
            if not wait:
                return command
            backoff_factor = 2
            if command.type == "BulkAddToTaskSetting":
                backoff_factor = 8
            command = command.wait(timeout=0, backoff_factor=backoff_factor)
            if task is None:
                task = self.get(task_id=task_id)
            if "addToTaskPayload" not in command.spec:
                raise exceptions.PlatformException(
                    error="400", message="addToTaskPayload key is missing in command response: {}".format(response)
                )
        else:
            raise exceptions.PlatformException(response)

        assert isinstance(task, entities.Task)
        return task

    # @_api_reference.add(path='/annotationtasks/{id}/removeFromTask', method='post')
    def remove_items(
        self,
        task: entities.Task = None,
        task_id=None,
        filters: entities.Filters = None,
        query=None,
        items=None,
        wait=True,
    ):
        """
        remove items from Task.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param dtlpy.entities.task.Task task: task object
        :param str task_id: the Id of the task
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param dict query: query to filter the items use it
        :param list items: list of items to add to the task
        :param bool wait: wait until remove items finish
        :return: True if success and an error if failed
        :rtype: bool

        **Examples**:

        .. code-block:: python

            dataset.tasks.remove_items(task= 'task_entity',
                                        items = [items])

        """
        if filters is None and items is None and query is None:
            raise exceptions.PlatformException("400", "Must provide either filters, query or items list")

        if task is None and task_id is None:
            raise exceptions.PlatformException("400", "Must provide either task or task_id")

        if query is None:
            if filters is None:
                if not isinstance(items, list):
                    items = [items]
                filters = entities.Filters(
                    field="id",
                    values=[item.id for item in items],
                    operator=entities.FiltersOperations.IN,
                    use_defaults=False,
                )
            query = filters.prepare()

        if task_id is None:
            task_id = task.id

        payload = {"query": "{}".format(json.dumps(query).replace("'", '"')), "asynced": wait}

        url = "{}/{}/removeFromTask".format(URL_PATH, task_id)

        success, response = self._client_api.gen_request(req_type="post", path=url, json_req=payload)

        if success:
            command = entities.Command.from_json(_json=response.json(), client_api=self._client_api)
            if not wait:
                return command
            command = command.wait(timeout=0)

            if "removeFromTaskId" not in command.spec:
                raise exceptions.PlatformException(
                    error="400", message="removeFromTaskId key is missing in command response: {}".format(response)
                )
        else:
            raise exceptions.PlatformException(response)
        return True

    def get_items(
        self,
        task_id: str = None,
        task_name: str = None,
        dataset: entities.Dataset = None,
        filters: entities.Filters = None,
        get_consensus_items: bool = False,
        task: entities.Task = None,
    ) -> entities.PagedEntities:
        """
        Get the task items to use in your code.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        If a filters param is provided, you will receive a PagedEntity output of the task items. If no filter is provided, you will receive a list of the items.

        :param str task_id: the id of the task
        :param str task_name: the name of the task
        :param bool get_consensus_items: get the items from the consensus assignment
        :param dtlpy.entities.Task task: task object
        :param dtlpy.entities.dataset.Dataset dataset: dataset object that refer to the task
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: list of the items or PagedEntity output of items
        :rtype: list or dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            dataset.tasks.get_items(task_id= 'task_id')
        """
        if task is None and task_id is None and task_name is None:
            raise exceptions.PlatformException("400", "Please provide either task_id or task_name")

        if task_id is None:
            if task is None:
                task = self.get(task_name=task_name)
            task_id = task.id

        if dataset is None and self._dataset is None:
            raise exceptions.PlatformException("400", "Please provide a dataset entity")
        if dataset is None:
            dataset = self._dataset

        if filters is None:
            filters = entities.Filters(use_defaults=False)
        filters.add(field="metadata.system.refs.id", values=[task_id], operator=entities.FiltersOperations.IN)

        if not get_consensus_items:
            if task is None:
                task = self.get(task_id=task_id)
            if task.metadata.get("system", dict()).get("consensusAssignmentId", None):
                filters.add(
                    field="metadata.system.refs.id",
                    values=task.metadata["system"]["consensusAssignmentId"],
                    operator=entities.FiltersOperations.NOT_EQUAL,
                )

        return dataset.items.list(filters=filters)

    def set_status(self, status: str, operation: str, task_id: str, item_ids: List[str]):
        """
        Update an item status within a task.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param str status: string the describes the status
        :param str operation: the status action need 'create' or 'delete'
        :param str task_id: the Id of the task
        :param list item_ids: List[str] id items ids
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            dataset.tasks.set_status(task_id= 'task_id', status='complete', operation='create')
        """
        url = "/assignments/items/tasks/{task_id}/status".format(task_id=task_id)
        payload = {
            "itemIds": item_ids,
            "statusPayload": {"operation": operation, "returnLastStatus": True, "status": status},
        }

        success, response = self._client_api.gen_request(req_type="post", path=url, json_req=payload)

        if not success:
            raise exceptions.PlatformException(response)
        if response.json() is not None:
            updated_items = set(response.json().keys())
            log_msg = "Items status was updated successfully."
            if len(updated_items) != len(item_ids):
                failed_items = set(item_ids).difference(updated_items)
                log_msg = "{success_count} out of TOTAL items were updated. The following items failed to update: {failed_items}".format(
                    success_count=len(updated_items), failed_items=failed_items
                )
            logger.info(msg=log_msg)
        return True

    def task_scores(self, task_id: str = None, page_offset: int = 0, page_size: int = 100):
        """
        Get all entities scores in a task.

        :param str task_id: the id of the task
        :param int page_offset: the page offset
        :param int page_size: the page size
        :return: page of the task scores

        **Example**:

        .. code-block:: python

            dataset.tasks.task_scores(task_id= 'task_id')
        """
        if task_id is None:
            raise exceptions.PlatformException("400", "Please provide task_id")

        url = "/scores/tasks/{task_id}?page={page_offset}&pageSize={page_size}".format(
            task_id=task_id, page_offset=page_offset, page_size=page_size
        )
        success, response = self._client_api.gen_request(req_type="get", path=url)

        if success:
            return response.json()
        else:
            raise exceptions.PlatformException(response)



import numpy as np
import logging
import time
import tqdm
import sys

from .. import exceptions, entities, miscellaneous
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')

MAX_SLEEP_TIME = 30


class Commands:
    """
    Service Commands repository
    """

    def __init__(self, client_api: ApiClient):
        self._client_api = client_api

    ############
    # entities #
    ############

    def list(self):
        """
        List of commands

        :return: list of commands
        """
        url = '/commands'

        # request
        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if not success:
            raise exceptions.PlatformException(response)

        commands = miscellaneous.List([entities.Command.from_json(_json=_json,
                                                                  client_api=self._client_api) for _json in
                                       response.json()])
        return commands

    def get(self,
            command_id: str = None,
            url: str = None
            ) -> entities.Command:
        """
        Get Service command object

        :param str command_id:
        :param str url: command url
        :return: Command object
        """
        if url is None:
            url_path = "/commands/{}".format(command_id)
        else:
            url_path = url.split('api/v1')[1]

        success, response = self._client_api.gen_request(req_type="get",
                                                         path=url_path)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Command.from_json(client_api=self._client_api,
                                          _json=response.json())

    def wait(self, command_id, timeout=0, step=None, url=None, backoff_factor=1, iteration_callback=None):
        """
        Wait for command to finish

        backoff_factor: A backoff factor to apply between attempts after the second try
        {backoff factor} * (2 ** ({number of total retries} - 1))
        seconds. If the backoff_factor is 1, then :func:`.sleep` will sleep
        for [0s, 2s, 4s, ...] between retries. It will never be longer
        than 30 sec

        :param str command_id: Command id to wait to
        :param int timeout: int, seconds to wait until TimeoutError is raised. if 0 - wait until done
        :param int step: int, seconds between polling
        :param str url: url to the command
        :param function iteration_callback: function to call on each iteration
        :param float backoff_factor: A backoff factor to apply between attempts after the second try
        :return: Command  object
        """

        elapsed = 0
        start = time.time()
        if timeout is None or timeout <= 0:
            timeout = np.inf

        command = None
        pbar = tqdm.tqdm(total=100,
                         disable=self._client_api.verbose.disable_progress_bar_command_progress,
                         file=sys.stdout,
                         desc='Command Progress')
        num_tries = 1
        while elapsed < timeout:
            command = self.get(command_id=command_id, url=url)
            if command.type == 'ExportDatasetAsJson':
                self._client_api.callbacks.run_on_event(event=self._client_api.callbacks.CallbackEvent.DATASET_EXPORT,
                                                        context=command.spec,
                                                        progress=command.progress)

            pbar.update(command.progress - pbar.n)
            if not command.in_progress():
                break
            elapsed = time.time() - start
            sleep_time = np.min([timeout - elapsed, backoff_factor * (2 ** num_tries), MAX_SLEEP_TIME])
            num_tries += 1
            logger.debug(f"Command {command.id} is running for {elapsed:.2f}[s]. Sleeping for {sleep_time:.2f}[s]")
            if iteration_callback is not None:
                try:
                    iteration_callback()
                except Exception as e:
                    logger.warning('iteration_callback failed: {}'.format(e.__str__()))
            time.sleep(sleep_time)
        pbar.close()
        if command is None:
            raise ValueError('Nothing to wait for')
        if elapsed >= timeout:
            raise TimeoutError("command wait() got timeout. id: {!r}, status: {}, progress {}%".format(
                command.id, command.status, command.progress))
        if command.status != entities.CommandsStatus.SUCCESS:
            raise exceptions.PlatformException(error='424',
                                               message="Command {!r} {}: '{}'".format(command.id,
                                                                                      command.status,
                                                                                      command.error))
        return command

    def abort(self, command_id: str):
        """
        Abort Command

        :param str command_id: command id
        :return:
        """
        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/commands/{}/abort'.format(command_id))

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        else:
            return entities.Command.from_json(_json=response.json(),
                                              client_api=self._client_api)



import os
from .service_runners import service_runner_paths


class PATHS:
    PACKAGE_FILENAME = 'package.json'
    GITIGNORE_FILENAME = 'package_gitignore'
    MAIN_FILENAME = 'main.py'
    MOCK_FILENAME = 'mock.json'
    MODEL_ADAPTER = 'model_adapter.py'
    PARTIAL_MAIN = 'main_partial.py'
    MODULE_A_FILENAME = 'first_module_class.py'
    MODULE_B_FILENAME = 'second_module_class.py'
    REQUIREMENTS_FILENAME = 'requirements.txt'
    APP_JSON_FILENAME = 'dataloop.json'

    ASSETS_PATH = os.path.dirname(__file__)
    ASSETS_MAIN_FILEPATH = os.path.join(ASSETS_PATH, MAIN_FILENAME)
    ASSETS_MOCK_FILEPATH = os.path.join(ASSETS_PATH, MOCK_FILENAME)
    ASSETS_PACKAGE_FILEPATH = os.path.join(ASSETS_PATH, PACKAGE_FILENAME)
    ASSETS_GITIGNORE_FILEPATH = os.path.join(ASSETS_PATH, GITIGNORE_FILENAME)
    ASSETS_MODEL_ADAPTER_FILEPATH = os.path.join(ASSETS_PATH, MODEL_ADAPTER)
    PARTIAL_MAIN_FILEPATH = os.path.join(ASSETS_PATH, PARTIAL_MAIN)


paths = PATHS()



import logging

from .. import entities, miscellaneous, exceptions, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Organizations:
    """
    Organizations Repository

    Read our `documentation <https://dataloop.ai/docs/org-setup>`_ to learn more about Organizations in the Dataloop platform.
    """

    def __init__(self, client_api: ApiClient):
        self._client_api = client_api

    def create(self, organization_json: dict) -> entities.Organization:
        """
        Create a new organization.

        **Prerequisites**: This method can only be used by a **superuser**.

        :param dict organization_json: json contain the Organization fields
        :return: Organization object
        :rtype: dtlpy.entities.organization.Organization
        """

        raise exceptions.PlatformException(error='2001', message='Method organizations.create() was removed')

    def list_groups(self, organization: entities.Organization = None,
                    organization_id: str = None,
                    organization_name: str = None):
        """
        List all organization groups (groups that were created within the organization).

        **Prerequisites**: You must be an organization *owner* to use this method.

        You must provide at least ONE of the following params: organization, organization_name, or organization_id.

        :param entities.Organization organization: Organization object
        :param str organization_id: Organization id
        :param str organization_name: Organization name
        :return: groups list
        :rtype: list

        **Example**:

        .. code-block:: python

            groups_list = dl.organizations.list_groups(organization_id='organization_id')
        """
        if organization is None and organization_id is None and organization_name is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        if organization is None:
            organization = self.get(organization_id=organization_id, organization_name=organization_name)

        url_path = '/orgs/{}/groups'.format(organization.id)

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)
        if not success:
            raise exceptions.PlatformException(response)

        groups = miscellaneous.List(response.json())
        return groups

    def list_integrations(self, organization: entities.Organization = None,
                          organization_id: str = None,
                          organization_name: str = None,
                          only_available=False):
        """
        List all organization integrations with external cloud storage.

        **Prerequisites**: You must be an organization *owner* to use this method.

        You must provide at least ONE of the following params: organization_id, organization_name, or organization.

        :param entities.Organization organization: Organization object
        :param str organization_id: Organization id
        :param str organization_name: Organization name
        :param bool only_available: if True list only the available integrations
        :return: integrations list
        :rtype: list

        **Example**:

        .. code-block:: python

            list_integrations = dl.organizations.list_integrations(organization='organization-entity',
                                                only_available=True)
        """
        if organization is None and organization_id is None and organization_name is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        if organization is None:
            organization = self.get(organization_id=organization_id, organization_name=organization_name)

        if only_available:
            url_path = '/orgs/{}/availableIntegrations'.format(organization.id)
        else:
            url_path = '/orgs/{}/integrations'.format(organization.id)

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)
        if not success:
            raise exceptions.PlatformException(response)

        available_integrations = miscellaneous.List(response.json())
        return available_integrations

    @_api_reference.add(path='/orgs/{orgId}/members', method='get')
    def list_members(self, organization: entities.Organization = None,
                     organization_id: str = None,
                     organization_name: str = None,
                     role: entities.MemberOrgRole = None):
        """
        List all organization members.

        **Prerequisites**: You must be an organization *owner* to use this method.

        You must provide at least ONE of the following params: organization_id, organization_name, or organization.

        :param entities.Organization organization: Organization object
        :param str organization_id: Organization id
        :param str organization_name: Organization name
        :param entities.MemberOrgRole role: MemberOrgRole.ADMIN, MemberOrgRole.OWNER, MemberOrgRole.MEMBER, MemberOrgRole.WORKER
        :return: projects list
        :rtype: list

        **Example**:

        .. code-block:: python

            list_members = dl.organizations.list_members(organization='organization-entity',
                                        role=dl.MemberOrgRole.MEMBER)
        """
        if organization is None and organization_id is None and organization_name is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        if organization is None:
            organization = self.get(organization_id=organization_id, organization_name=organization_name)

        url_path = '/orgs/{}/members'.format(organization.id)

        if role is not None and role not in list(entities.MemberOrgRole):
            raise ValueError('Unknown role {!r}, role must be one of: {}'.format(role,
                                                                                 ', '.join(
                                                                                     list(entities.MemberOrgRole))))

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)
        if not success:
            raise exceptions.PlatformException(response)

        members = miscellaneous.List(
            [entities.User.from_json(_json=user, client_api=self._client_api, project=None) for user in
             response.json()])

        if role is not None:
            members = [member for member in members if member.role == role]

        return members

    @_api_reference.add(path='/orgs', method='get')
    def list(self) -> miscellaneous.List[entities.Organization]:
        """
        Lists all the organizations in Dataloop.

        **Prerequisites**: You must be a **superuser** to use this method.

        :return: List of Organization objects
        :rtype: list

        **Example**:

        .. code-block:: python

            organizations = dl.organizations.list()
        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/orgs')

        if success:
            pool = self._client_api.thread_pools(pool_name='entity.create')
            organization_json = response.json()
            jobs = [None for _ in range(len(organization_json))]
            # return triggers list
            for i_organization, organization in enumerate(organization_json):
                jobs[i_organization] = pool.submit(entities.Organization._protected_from_json,
                                                   **{'client_api': self._client_api,
                                                      '_json': organization})

            # get all results
            results = [j.result() for j in jobs]
            # log errors
            _ = [logger.warning(r[1]) for r in results if r[0] is False]
            # return good jobs
            organization = miscellaneous.List([r[1] for r in results if r[0] is True])
        else:
            logger.error('Platform error getting organization')
            raise exceptions.PlatformException(response)
        return organization

    @_api_reference.add(path='/orgs/{orgId}', method='get')
    def get(self,
            organization_id: str = None,
            organization_name: str = None,
            fetch: bool = None) -> entities.Organization:
        """
        Get Organization object to be able to use it in your code.

        **Prerequisites**: You must be a **superuser** to use this method.

        You must provide at least ONE of the following params: organization_name or organization_id.

        :param str organization_id: optional - search by id
        :param str organization_name: optional - search by name
        :param fetch: optional - fetch entity from platform, default taken from cookie
        :return: Organization object
        :rtype: dtlpy.entities.organization.Organization

        **Example**:

        .. code-block:: python

            org = dl.organizations.get(organization_id='organization_id')
        """
        if organization_name is None and organization_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        if fetch is None:
            fetch = self._client_api.fetch_entities

        if fetch:
            if organization_id is not None:
                success, response = self._client_api.gen_request(req_type='get',
                                                                 path='/orgs/{}'.format(organization_id))
                if not success:
                    raise exceptions.PlatformException(response)
                organization = entities.Organization.from_json(
                    client_api=self._client_api,
                    _json=response.json()
                )
            else:
                organizations = self.list()
                organization = [organization for organization in organizations if
                                organization.name == organization_name]
                if not organization:
                    # list is empty
                    raise exceptions.PlatformException(error='404',
                                                       message='organization not found. Name: {}'.format(
                                                           organization_name))
                    # project = None
                elif len(organization) > 1:
                    # more than one matching project
                    raise exceptions.PlatformException(
                        error='404',
                        message='More than one project with same name. Please "get" by id')
                else:
                    organization = organization[0]
        else:
            organization = entities.Organization.from_json(
                _json={'id': organization_id,
                       'name': organization_name},
                client_api=self._client_api)

        return organization

    def update(self, plan: str,
               organization: entities.Organization = None,
               organization_id: str = None,
               organization_name: str = None) -> entities.Organization:
        """
        Update an organization.

        **Prerequisites**: You must be a **superuser** to update an organization.

        You must provide at least ONE of the following params: organization, organization_name, or organization_id.

        :param str plan: OrganizationsPlans.FREEMIUM, OrganizationsPlans.PREMIUM
        :param entities.Organization organization: Organization object
        :param str organization_id: Organization id
        :param str organization_name: Organization name
        :return: organization object
        :rtype: dtlpy.entities.organization.Organization

        **Example**:

        .. code-block:: python

            org = dl.organizations.update(organization='organization-entity',
                                    plan=dl.OrganizationsPlans.FREEMIUM)
        """
        if organization is None and organization_id is None and organization_name is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        if organization is None:
            organization = self.get(organization_id=organization_id, organization_name=organization_name)

        if plan not in list(entities.OrganizationsPlans):
            raise ValueError('Unknown role {!r}, role must be one of: {}'.format(plan,
                                                                                 ', '.join(list(
                                                                                     entities.OrganizationsPlans))))
        payload = {'plan': plan}
        url_path = '/orgs/{}/plan'.format(organization.id)
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=payload)
        if success:
            return organization
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/orgs/{orgId}/members', method='post')
    def add_member(self, email: str,
                   role: entities.MemberOrgRole = entities.MemberOrgRole.MEMBER,
                   organization_id: str = None,
                   organization_name: str = None,
                   organization: entities.Organization = None):
        """
        Add members to your organization. Read about members and groups `here <https://dataloop.ai/docs/org-members-groups>`_.

        **Prerequisities**: To add members to an organization, you must be an *owner* in that organization.

        You must provide at least ONE of the following params: organization, organization_name, or organization_id.

        :param str email: the member's email
        :param str role: MemberOrgRole.ADMIN, MemberOrgRole.OWNER, MemberOrgRole.MEMBER, MemberOrgRole.WORKER
        :param str organization_id: Organization id
        :param str organization_name: Organization name
        :param entities.Organization organization: Organization object
        :return: True if successful or error if unsuccessful
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = dl.organizations.add_member(email='user@domain.com',
                                        organization_id='organization_id',
                                        role=dl.MemberOrgRole.MEMBER)
        """

        if organization is None and organization_id is None and organization_name is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        if organization is None:
            organization = self.get(organization_id=organization_id, organization_name=organization_name)

        if not isinstance(email, list):
            email = [email]

        if role not in list(entities.MemberOrgRole):
            raise ValueError('Unknown role {!r}, role must be one of: {}'.format(role,
                                                                                 ', '.join(
                                                                                     list(entities.MemberOrgRole))))

        url_path = '/orgs/{}/members'.format(organization.id)
        payload = {"emails": email, 'role': role}
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        else:
            return True

    @_api_reference.add(path='/orgs/{orgId}/members/{memberId}', method='delete')
    def delete_member(self, user_id: str,
                      organization_id: str = None,
                      organization_name: str = None,
                      organization: entities.Organization = None,
                      sure: bool = False,
                      really: bool = False) -> bool:
        """
        Delete member from the Organization.

        **Prerequisites**: Must be an organization *owner* to delete members.

        You must provide at least ONE of the following params: organization_id, organization_name, organization.

        :param str user_id: user id
        :param str organization_id: Organization id
        :param str organization_name: Organization name
        :param entities.Organization organization: Organization object
        :param bool sure: Are you sure you want to delete?
        :param bool really: Really really sure?
        :return: True if success and error if not
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = dl.organizations.delete_member(user_id='user_id',
                                            organization_id='organization_id',
                                            sure=True,
                                            really=True)
        """
        if sure and really:
            if organization is None and organization_id is None and organization_name is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='Must provide an identifier in inputs')

            if organization is None:
                organization = self.get(organization_id=organization_id, organization_name=organization_name)

            url_path = '/orgs/{}/members/{}'.format(organization.id, user_id)
            success, response = self._client_api.gen_request(req_type='delete',
                                                             path=url_path)
            if not success:
                raise exceptions.PlatformException(response)
            else:
                return True
        else:
            raise exceptions.PlatformException(
                error='403',
                message='Cant delete member from SDK. Please login to platform to delete')

    @_api_reference.add(path='/orgs/{orgId}/members', method='patch')
    def update_member(self, email: str,
                      role: entities.MemberOrgRole = entities.MemberOrgRole.MEMBER,
                      organization_id: str = None,
                      organization_name: str = None,
                      organization: entities.Organization = None):
        """
        Update member role.

        **Prerequisites**: You must be an organization *owner* to update a member's role.

        You must provide at least ONE of the following params: organization, organization_name, or organization_id.

        :param str email: the member's email
        :param str role: MemberOrgRole.ADMIN, MemberOrgRole.OWNER, MemberOrgRole.MEMBER, MemberOrgRole.WORKER
        :param str organization_id: Organization id
        :param str organization_name: Organization name
        :param entities.Organization organization: Organization object
        :return: json of the member fields
        :rtype: dict

        **Example**:

        .. code-block:: python

            member_json = dl.organizations.update_member(email='user@domain.com',
                                            organization_id='organization_id',
                                             role=dl.MemberOrgRole.MEMBER)
        """
        if organization is None and organization_id is None and organization_name is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        if organization is None:
            organization = self.get(organization_id=organization_id, organization_name=organization_name)

        url_path = '/orgs/{}/members'.format(organization.id)
        payload = dict(role=role, email=email)

        if role not in list(entities.MemberOrgRole):
            raise ValueError('Unknown role {!r}, role must be one of: {}'.format(role,
                                                                                 ', '.join(
                                                                                     list(entities.MemberOrgRole))))

        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    def cache_action(self,
                     organization_id: str = None,
                     organization_name: str = None,
                     organization: entities.Organization = None,
                     mode=entities.CacheAction.APPLY,
                     pod_type=entities.PodType.SMALL):
        """
        Add or remove Cache for the org

        **Prerequisites**: You must be an organization *owner*

        You must provide at least ONE of the following params: organization, organization_name, or organization_id.

        :param str organization_id: Organization id
        :param str organization_name: Organization name
        :param entities.Organization organization: Organization object
        :param str mode: dl.CacheAction.APPLY or dl.CacheAction.DESTROY
        :param entities.PodType pod_type:  dl.PodType.SMALL, dl.PodType.MEDIUM, dl.PodType.HIGH
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = dl.organizations.enable_cache(organization_id='organization_id',
                                          mode=dl.CacheAction.APPLY)
        """
        if organization is None and organization_id is None and organization_name is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        if organization is None:
            organization = self.get(organization_id=organization_id, organization_name=organization_name)

        return organization.cache_action(mode=mode, pod_type=pod_type)



class ApiReference(object):

    def __init__(self):
        self.references = {'paths': dict()}

    def add(self, path, method):
        def wrapper(f):
            if path not in self.references['paths']:
                self.references['paths'][path] = dict()
            if method not in self.references['paths'][path]:
                self.references['paths'][path][method] = dict()

            docstring = f.__doc__
            docstring = docstring.split('\n')
            in_code = False

            code_list = list()
            docs_list = list()
            for i_line in range(len(docstring)):
                if '.. code-block::' in docstring[i_line]:
                    in_code = True
                    continue
                if '**Example**' in docstring[i_line]:
                    continue
                if docstring[i_line].startswith('    ') and not docstring[i_line].startswith('        '):
                    in_code = False
                if in_code:
                    code_list.append('{}'.format(docstring[i_line].strip()))
                else:
                    docs_list.append('# {}'.format(docstring[i_line].strip()))
                # assert False
            docstring = '\n'.join(code_list + docs_list)
            self.references['paths'][path][method]['x-codeSamples'] = [{"lang": "Python",
                                                                        "source": docstring}]
            return f

        return wrapper


api_reference = ApiReference()



import logging
from typing import Union

from .. import exceptions, entities, ApiClient

logger = logging.getLogger(name='dtlpy')
BASE_URL = '/settings'


class Settings:
    def __init__(
            self,
            client_api: ApiClient,
            project: entities.Project = None,
            dataset: entities.Dataset = None,
            org: entities.Organization = None,
            task: entities.Task = None,
            resource=None,
            resource_type=None
    ):
        self._client_api = client_api
        self._org = org
        self._project = project
        self._dataset = dataset
        self._task = task
        self._resource = resource
        self._resource_type = resource_type

    ###########
    # methods #
    ###########

    @staticmethod
    def get_constructor(res: dict):
        constructor = entities.Setting.from_json
        return constructor

    def _build_entities_from_response(self, response_items):
        settings = list()
        for _, setting in enumerate(response_items):
            settings.append(
                self.get_constructor(setting)(
                    client_api=self._client_api,
                    _json=setting,
                    project=self._project
                )
            )

        return settings

    def _build_settings(self,
                        setting_name: str = None,
                        setting_value: entities.SettingsTypes = None,
                        setting_value_type: entities.SettingsValueTypes = None,
                        setting_default_value=None
                        ):
        if self._resource is None:
            raise exceptions.PlatformException('400', 'Must have resource')
        setting = entities.Setting(
            name=setting_name,
            value=setting_value,
            value_type=setting_value_type,
            section_name=entities.SettingsSectionNames.SDK,
            default_value=setting_default_value,
            scope=entities.SettingScope(type=self._resource_type,
                                        id=self._resource.id,
                                        role=entities.Role.ALL,
                                        prevent_override=False,
                                        visible=True),
        )
        return setting

    def create(self,
               setting: entities.Setting = None,
               setting_name: str = None,
               setting_value=None,
               setting_value_type: entities.SettingsValueTypes = None,
               setting_default_value=None,
               ) -> entities.Setting:
        """
        Create a new setting

        :param Setting setting: setting entity
        :param str setting_name: the setting name
        :param setting_value: the setting value
        :param SettingsValueTypes setting_value_type: the setting type dl.SettingsValueTypes
        :param setting_default_value: the setting default value
        :return: setting entity
        """
        if sum([1 for param in [setting_name, setting_value, setting_value_type] if param is None]) > 0 \
                and setting is None:
            raise exceptions.PlatformException('400', 'Must provide setting object or'
                                                      'setting_name, setting_value and setting_value_type')

        if setting is None:
            setting = self._build_settings(setting_name=setting_name,
                                           setting_value=setting_value,
                                           setting_value_type=setting_value_type,
                                           setting_default_value=setting_default_value)

        success, response = self._client_api.gen_request(
            req_type='post',
            path='{}'.format(
                BASE_URL
            ),
            json_req=setting.to_json()
        )

        if success:
            _json = response.json()
            constructor = self.get_constructor(_json)
        else:
            raise exceptions.PlatformException(response)

        return constructor(
            _json=_json,
            client_api=self._client_api,
            project=self._project,
            org=self._org
        )

    def update(self,
               setting: entities.BaseSetting = None,
               setting_id: str = None,
               setting_name: str = None,
               setting_value=None,
               setting_default_value=None,
               ) -> entities.Setting:
        """
        Update a setting

        :param Setting setting: setting entity
        :param str setting_id: the setting id
        :param str setting_name: the setting name
        :param setting_value: the setting value
        :param setting_default_value: the setting default value
        :return: setting entity
        """
        if sum([1 for param in [setting_name, setting_value] if param is None]) > 0 \
                and setting is None:
            raise exceptions.PlatformException('400', 'Must provide setting object or'
                                                      'setting_name, setting_value and setting_value_type')

        if setting is None:
            setting = self.get(setting_id=setting_id)
            setting.name = setting_name
            setting.value = setting_value
            if setting_default_value is not None:
                setting.default_value = setting_default_value

        patch = setting.to_json()
        patch.pop('id')
        patch.pop('name')
        success, response = self._client_api.gen_request(
            req_type='patch',
            path='{}/{}'.format(
                BASE_URL,
                setting.id
            ),
            json_req=patch
        )

        if success:
            _json = response.json()
            constructor = self.get_constructor(_json)
        else:
            raise exceptions.PlatformException(response)

        return constructor(
            _json=_json,
            client_api=self._client_api,
            project=self._project,
            org=self._org
        )

    def delete(self, setting_id: str) -> bool:
        """
        Delete a setting

        :param str setting_id: the setting id
        :return: True if success exceptions if not
        """
        success, response = self._client_api.gen_request(
            req_type='delete',
            path='{}/{}'.format(
                BASE_URL,
                setting_id
            )
        )

        if success:
            return True
        else:
            raise exceptions.PlatformException(response)

    def get(self, setting_name: str = None, setting_id: str = None) -> entities.Setting:
        """
        Get a setting by id

        :param str setting_name: the setting name
        :param str setting_id: the setting id
        :return: setting entity
        """
        if setting_id is not None:
            success, response = self._client_api.gen_request(
                req_type='get',
                path='{}/{}'.format(
                    BASE_URL,
                    setting_id
                )
            )

            if success:
                _json = response.json()
                constructor = self.get_constructor(_json)
                setting = constructor(
                    _json=_json,
                    client_api=self._client_api,
                    project=self._project,
                    org=self._org
                )
            else:
                raise exceptions.PlatformException(response)
        elif setting_name is not None:
            if not isinstance(setting_name, str):
                raise exceptions.PlatformException(
                    error='400',
                    message='setting_name must be strings')
            settings = self.list(
                filters=entities.Filters(field='name', values=setting_name, resource=entities.FiltersResource.SETTINGS))
            if len(settings.items) > 1:
                raise exceptions.PlatformException(
                    error='404',
                    message='More than one setting with same name. Please "get" by id')
            elif len(settings.items) == 0:
                raise exceptions.PlatformException('404', 'setting not found')
            else:
                setting = settings.items[0]
        else:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide at least ONE of the following params: setting_id, setting_name.')
        return setting

    def _list(self, filters: entities.Filters):
        success, response = self._client_api.gen_request(
            req_type='post',
            path='{}/query'.format(BASE_URL),
            json_req=filters.prepare()
        )
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List settings

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.SETTINGS)
            filters.sort_by(entities.FiltersOrderByDirection.ASCENDING)

        if filters.resource != entities.FiltersResource.SETTINGS:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.SETTINGS . Got: {!r}'.format(filters.resource))

        if self._project is not None:
            filters.add(field='scope.id', values=self._project.id)

        paged = entities.PagedEntities(
            items_repository=self,
            filters=filters,
            page_offset=filters.page,
            page_size=filters.page_size,
            project_id=self._project.id if self._project is not None else None,
            client_api=self._client_api
        )

        paged.get_page()

        return paged

    def resolve(
            self,
            user_email: str,
            org_id: str = None,
            project_id: str = None,
            dataset_id: str = None,
            task_id: str = None,
    ):
        """
        return all the settings that relevant to the provider params

        :param str user_email: user email
        :param str org_id: org id
        :param str project_id: project id
        :param str dataset_id: dataset id
        :param str task_id: task id
        """
        payload = {
            'userId': user_email
        }

        if self._project:
            payload['projectId'] = self._project.id
        elif project_id:
            payload['projectId'] = project_id

        if self._org:
            payload['orgId'] = self._org.id
        elif org_id:
            payload['orgId'] = org_id

        if self._dataset:
            payload['datasetId'] = self._dataset.id
        elif dataset_id:
            payload['datasetId'] = dataset_id

        if self._task:
            payload['taskId'] = self._task.id
        elif task_id:
            payload['taskId'] = task_id

        success, response = self._client_api.gen_request(
            req_type='post',
            path='{}/resolve'.format(BASE_URL),
            json_req=payload
        )

        if success:
            _json = response.json()
            return self._build_entities_from_response(response_items=_json)
        else:
            raise exceptions.PlatformException(response)



import attr
import traceback
from collections import namedtuple
from enum import Enum

from .. import repositories, entities
from ..services.api_client import ApiClient


@attr.s
class Feature(entities.BaseEntity):
    """
    Com entity
    """
    # platform
    id = attr.ib()
    entity_id = attr.ib()
    url = attr.ib(repr=False)
    created_at = attr.ib()
    feature_set_id = attr.ib()
    version = attr.ib()
    value = attr.ib()
    parent_id = attr.ib()
    project_id = attr.ib()
    org_id = attr.ib()
    creator = attr.ib()

    # sdk
    _client_api = attr.ib(type=ApiClient, repr=False)
    _repositories = attr.ib(repr=False)

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['features', 'features_sets'])
        features_repo = repositories.Features(client_api=self._client_api)
        r = reps(features=features_repo,
                 features_sets=repositories.FeatureSets(client_api=self._client_api))
        return r

    @property
    def features(self):
        assert isinstance(self._repositories.features, repositories.Features)
        return self._repositories.features

    @staticmethod
    def _protected_from_json(_json, client_api, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json:
        :param client_api:
        :return:
        """
        try:
            feature = Feature.from_json(_json=_json,
                                        client_api=client_api,
                                        is_fetched=is_fetched)
            status = True
        except Exception:
            feature = traceback.format_exc()
            status = False
        return status, feature

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Feature entity object from a json

        :param is_fetched: is Entity fetched from Platform
        :param _json: _json response from host
        :param client_api: client_api
        :return: Feature object
        """
        inst = cls(
            id=_json.get('id', None),
            feature_set_id=_json.get('featureSetId', None),
            entity_id=_json.get('entityId', None),
            url=_json.get('url', None),
            project_id=_json.get('project', None),
            created_at=_json.get('createdAt', None),
            version=_json.get('version', None),
            value=_json.get('value', None),
            parent_id=_json.get('parentId', None),
            client_api=client_api,
            org_id=_json.get('org', None),
            creator=_json.get('creator', None),
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """

        _json = {'createdAt': self.created_at,
                 'entityId': self.entity_id,
                 'id': self.id,
                 'featureSetId': self.feature_set_id,
                 'url': self.url,
                 'project': self.project_id,
                 'creator': self.creator,
                 'version': self.version,
                 'value': self.value,
                 }
        if self.parent_id is not None:
            _json['parentId'] = self.parent_id
        if self.org_id is not None:
            _json['org'] = self.org_id
        return _json

    def delete(self):
        """
        Delete Feature Vector object

        :return: True
        """
        return self.features.delete(feature_id=self.id)



from .. import entities, exceptions
from ..services.api_client import ApiClient


class Nodes(list):

    def __init__(self, client_api: ApiClient, pipeline: entities.Pipeline = None, nodes=None):
        if nodes is None:
            nodes = []
        self._client_api = client_api
        self._pipeline = pipeline
        for node in nodes:
            self.add(node)

    def add(self, node: entities.PipelineNode) -> entities.PipelineNode:
        """
        Add a node to the nodes list

        :param PipelineNode node: node to add
        :return: node that add
        """
        if node not in self:
            self.append(node)
            node._pipeline = self._pipeline
            if not self._pipeline.start_nodes:
                self._pipeline.set_start_node(node=node)
        return node

    def get(self, node_name: str = None, node_id: str = None) -> entities.PipelineNode:
        """
        Get a node from the nodes list by name

        :param str node_name: the node name
        :param str node_id: the node id
        :return: the result node
        """
        if node_id is None and node_name is None:
            raise exceptions.PlatformException('400',
                                               'Must provide node_id or node_name')
        for node in self:
            if node_name == node.name or node_id == node.node_id:
                return node
        return None

    def remove(self, node_name) -> bool:
        """
        Remove a node from the nodes list by name

        :param str node_name: the node name
        :return: True if success
        """
        if not isinstance(node_name, str):
            raise ValueError('node name must be string')

        # get the wanted node
        node = self.get(node_name=node_name)
        if node:
            copy_conn = self._pipeline.connections.copy()
            # remove all node connections
            for conn in copy_conn:
                if node.node_id == conn.source.node_id or node.node_id == conn.target.node_id:
                    self._pipeline.connections.remove(conn)
            # remove the node
            for node_index in range(len(self)):
                if self[node_index].node_id == node.node_id:
                    self.pop(node_index)
                    # remove the node from the start_nodes if it exists
                    # if the node is the root node set the first node as the start node
                    for n in self._pipeline.start_nodes:
                        if n['nodeId'] == node.node_id:
                            # check if still have nodes in the pipeline nodes list
                            if self:
                                if n['type'] == 'root':
                                    self._pipeline.set_start_node(self[0])
                                else:
                                    self._pipeline.start_nodes.remove(n)
                            else:
                                self._pipeline.start_nodes = []
                    return True
        return False



def main():
    """
    This is an example how to upload files and annotations to Dataloop platform.
    Image folder contains the images to upload.
    Annotations folder contains json file of the annotations. Same name as the image.
    We read the images one by one and create the Dataloop annotations using the annotation builder.
    Finally, we upload both the image and the matching annotations

    :return:
    """
    import json
    import os
    import dtlpy as dl

    # Get project and dataset
    project = dl.projects.get(project_name='Yachts')
    dataset = project.datasets.get(dataset_name='Open Seas')

    images_folder = '/home/local/images'
    annotations_folder = '/home/local/annotations'

    for img_filename in os.listdir(images_folder):
        # get the matching annotations json
        _, ext = os.path.splitext(img_filename)
        ann_filename = os.path.join(annotations_folder, img_filename.replace(ext, '.json'))
        img_filename = os.path.join(images_folder, img_filename)

        # Upload or get annotations from platform (if already exists)
        item = dataset.items.upload(local_path=img_filename,
                                    remote_path='/in_storm',
                                    overwrite=False)
        assert isinstance(item, dl.Item)

        # read annotations from file
        with open(ann_filename, 'r') as f:
            annotations = json.load(f)

        # create a Builder instance and add all annotations to it
        builder = item.annotations.builder()
        for annotation in annotations:
            # line format if 4 points of bbox
            # this is where you need to update according to your annotation format
            label_id = annotation['label']
            left = annotation['left']
            top = annotation['top']
            right = annotation['right']
            bottom = annotation['bottom']
            builder.add(annotation_definition=dl.Box(top=top,
                                                     left=left,
                                                     bottom=bottom,
                                                     right=right,
                                                     label=str(label_id)))

        # upload annotations
        item.annotations.upload(builder)
